"""AiiDA workflow components: WorkGraph."""
from __future__ import annotations

import asyncio
import collections.abc
import functools
import logging
import typing as t

from plumpy import process_comms
from plumpy.persistence import auto_persist
from plumpy.process_states import Continue, Wait
from plumpy.workchains import _PropagateReturn
import kiwipy

from aiida.common import exceptions
from aiida.common.extendeddicts import AttributeDict
from aiida.common.lang import override
from aiida import orm
from aiida.orm import load_node, Node, ProcessNode, WorkChainNode
from aiida.orm.utils.serialize import deserialize_unsafe, serialize

from aiida.engine.processes.exit_code import ExitCode
from aiida.engine.processes.process import Process

from aiida.engine.processes.workchains.awaitable import (
    Awaitable,
    AwaitableAction,
    AwaitableTarget,
    construct_awaitable,
)
from aiida.engine.processes.workchains.workchain import Protect, WorkChainSpec
from aiida.engine import run_get_node
from aiida_workgraph.utils import create_and_pause_process
from aiida_workgraph.task import Task
from aiida_workgraph.utils import get_nested_dict, update_nested_dict
from aiida_workgraph.executors.monitors import monitor

if t.TYPE_CHECKING:
    from aiida.engine.runners import Runner  # pylint: disable=unused-import

__all__ = "WorkGraph"


MAX_NUMBER_AWAITABLES_MSG = "The maximum number of subprocesses has been reached: {}. Cannot launch the job: {}."


@auto_persist("_awaitables")
class WorkGraphEngine(Process, metaclass=Protect):
    """The `WorkGraph` class is used to construct workflows in AiiDA."""

    # used to create a process node that represents what happened in this process.
    _node_class = WorkChainNode
    _spec_class = WorkChainSpec
    _CONTEXT = "CONTEXT"

    def __init__(
        self,
        inputs: dict | None = None,
        logger: logging.Logger | None = None,
        runner: "Runner" | None = None,
        enable_persistence: bool = True,
    ) -> None:
        """Construct a WorkGraph instance.

        :param inputs: work graph inputs
        :param logger: aiida logger
        :param runner: work graph runner
        :param enable_persistence: whether to persist this work graph

        """

        super().__init__(inputs, logger, runner, enable_persistence=enable_persistence)

        self._awaitables: list[Awaitable] = []
        self._context = AttributeDict()

    @classmethod
    def define(cls, spec: WorkChainSpec) -> None:
        super().define(spec)
        spec.input("input_file", valid_type=orm.SinglefileData, required=False)
        spec.input_namespace(
            "wg", dynamic=True, required=False, help="WorkGraph inputs"
        )
        spec.input_namespace("input_tasks", dynamic=True, required=False)
        spec.exit_code(2, "ERROR_SUBPROCESS", message="A subprocess has failed.")

        spec.outputs.dynamic = True

        spec.output(
            "execution_count",
            valid_type=orm.Int,
            required=False,
            help="The number of time the WorkGraph runs.",
        )
        #
        spec.exit_code(
            201, "UNKNOWN_MESSAGE_TYPE", message="The message type is unknown."
        )
        spec.exit_code(202, "UNKNOWN_TASK_TYPE", message="The task type is unknown.")
        #
        spec.exit_code(
            301,
            "OUTPUS_NOT_MATCH_RESULTS",
            message="The outputs of the process do not match the results.",
        )
        spec.exit_code(
            302,
            "TASK_FAILED",
            message="Some of the tasks failed.",
        )
        spec.exit_code(
            303,
            "TASK_NON_ZERO_EXIT_STATUS",
            message="Some of the tasks exited with non-zero status.",
        )

    @property
    def ctx(self) -> AttributeDict:
        """Get the context."""
        return self._context

    @override
    def save_instance_state(
        self, out_state: t.Dict[str, t.Any], save_context: t.Any
    ) -> None:
        """Save instance state.

        :param out_state: state to save in

        :param save_context:
        :type save_context: :class:`!plumpy.persistence.LoadSaveContext`

        """
        super().save_instance_state(out_state, save_context)
        # Save the context
        out_state[self._CONTEXT] = self.ctx

    @override
    def load_instance_state(
        self, saved_state: t.Dict[str, t.Any], load_context: t.Any
    ) -> None:
        super().load_instance_state(saved_state, load_context)
        # Load the context
        self._context = saved_state[self._CONTEXT]
        self._temp = {"awaitables": {}}

        self.set_logger(self.node.logger)

        if self._awaitables:
            # For the "ascyncio.tasks.Task" awaitable, because there are only in-memory,
            # we need to reset the tasks and so that they can be re-run again.
            should_resume = False
            for awaitable in self._awaitables:
                if awaitable.target == "asyncio.tasks.Task":
                    self._resolve_awaitable(awaitable, None)
                    self.report(f"reset awaitable task: {awaitable.key}")
                    self.reset_task(awaitable.key)
                    should_resume = True
            if should_resume:
                self._update_process_status()
                self.resume()
            # For other awaitables, because they exist in the db, we only need to re-register the callbacks
            self.ctx._awaitable_actions = []
            self._action_awaitables()

    def _resolve_nested_context(self, key: str) -> tuple[AttributeDict, str]:
        """
        Returns a reference to a sub-dictionary of the context and the last key,
        after resolving a potentially segmented key where required sub-dictionaries are created as needed.

        :param key: A key into the context, where words before a dot are interpreted as a key for a sub-dictionary
        """
        ctx = self.ctx
        ctx_path = key.split(".")

        for index, path in enumerate(ctx_path[:-1]):
            try:
                ctx = ctx[path]
            except KeyError:  # see below why this is the only exception we have to catch here
                ctx[
                    path
                ] = AttributeDict()  # create the sub-dict and update the context
                ctx = ctx[path]
                continue

            # Notes:
            # * the first ctx (self.ctx) is guaranteed to be an AttributeDict, hence the post-"dereference" checking
            # * the values can be many different things: on insertion they are either AtrributeDict, List or Awaitables
            #   (subclasses of AttributeDict) but after resolution of an Awaitable this will be the value itself
            # * assumption: a resolved value is never a plain AttributeDict, on the other hand if a resolved Awaitable
            #   would be an AttributeDict we can append things to it since the order of tasks is maintained.
            if type(ctx) != AttributeDict:  # pylint: disable=C0123
                raise ValueError(
                    f"Can not update the context for key `{key}`: "
                    f' found instance of `{type(ctx)}` at `{".".join(ctx_path[:index + 1])}`, expected AttributeDict'
                )

        return ctx, ctx_path[-1]

    def _insert_awaitable(self, awaitable: Awaitable) -> None:
        """Insert an awaitable that should be terminated before before continuing to the next step.

        :param awaitable: the thing to await
        """
        ctx, key = self._resolve_nested_context(awaitable.key)

        # Already assign the awaitable itself to the location in the context container where it is supposed to end up
        # once it is resolved. This is especially important for the `APPEND` action, since it needs to maintain the
        # order, but the awaitables will not necessarily be resolved in the order in which they are added. By using the
        # awaitable as a placeholder, in the `_resolve_awaitable`, it can be found and replaced by the resolved value.
        if awaitable.action == AwaitableAction.ASSIGN:
            ctx[key] = awaitable
        elif awaitable.action == AwaitableAction.APPEND:
            ctx.setdefault(key, []).append(awaitable)
        else:
            raise AssertionError(f"Unsupported awaitable action: {awaitable.action}")

        self._awaitables.append(
            awaitable
        )  # add only if everything went ok, otherwise we end up in an inconsistent state
        self._update_process_status()

    def _resolve_awaitable(self, awaitable: Awaitable, value: t.Any) -> None:
        """Resolve an awaitable.

        Precondition: must be an awaitable that was previously inserted.

        :param awaitable: the awaitable to resolve
        :param value: the value to assign to the awaitable
        """
        ctx, key = self._resolve_nested_context(awaitable.key)

        if awaitable.action == AwaitableAction.ASSIGN:
            ctx[key] = value
        elif awaitable.action == AwaitableAction.APPEND:
            # Find the same awaitable inserted in the context
            container = ctx[key]
            for index, placeholder in enumerate(container):
                if (
                    isinstance(placeholder, Awaitable)
                    and placeholder.pk == awaitable.pk
                ):
                    container[index] = value
                    break
            else:
                raise AssertionError(
                    f"Awaitable `{awaitable.pk} was not in `ctx.{awaitable.key}`"
                )
        else:
            raise AssertionError(f"Unsupported awaitable action: {awaitable.action}")

        awaitable.resolved = True
        # remove awaitabble from the list
        self._awaitables = [a for a in self._awaitables if a.pk != awaitable.pk]

        if not self.has_terminated():
            # the process may be terminated, for example, if the process was killed or excepted
            # then we should not try to update it
            self._update_process_status()

    @Protect.final
    def to_context(self, **kwargs: Awaitable | ProcessNode) -> None:
        """Add a dictionary of awaitables to the context.

        This is a convenience method that provides syntactic sugar, for a user to add multiple intersteps that will
        assign a certain value to the corresponding key in the context of the work graph.
        """
        for key, value in kwargs.items():
            awaitable = construct_awaitable(value)
            awaitable.key = key
            self._insert_awaitable(awaitable)

    def _update_process_status(self) -> None:
        """Set the process status with a message accounting the current sub processes that we are waiting for."""
        if self._awaitables:
            status = f"Waiting for child processes: {', '.join([str(_.pk) for _ in self._awaitables])}"
            self.node.set_process_status(status)
        else:
            self.node.set_process_status(None)

    @override
    def run(self) -> t.Any:
        self.setup()
        return self._do_step()

    def _do_step(self) -> t.Any:
        """Execute the next step in the workgraph and return the result.

        If any awaitables were created, the process will enter in the Wait state,
        otherwise it will go to Continue.
        """
        # we will not remove the awaitables here,
        # we resume the workgraph in the callback function even
        # there are some awaitables left
        # self._awaitables = []
        result: t.Any = None

        try:
            self.continue_workgraph()
        except _PropagateReturn as exception:
            finished, result = True, exception.exit_code
        else:
            finished, result = self.is_workgraph_finished()

        # If the workgraph is finished or the result is an ExitCode, we exit by returning
        if finished:
            if isinstance(result, ExitCode):
                return result
            else:
                return self.finalize()

        if self._awaitables:
            return Wait(self._do_step, "Waiting before next step")

        return Continue(self._do_step)

    def _store_nodes(self, data: t.Any) -> None:
        """Recurse through a data structure and store any unstored nodes that are found along the way

        :param data: a data structure potentially containing unstored nodes
        """
        if isinstance(data, Node) and not data.is_stored:
            data.store()
        elif isinstance(data, collections.abc.Mapping):
            for _, value in data.items():
                self._store_nodes(value)
        elif isinstance(data, collections.abc.Sequence) and not isinstance(data, str):
            for value in data:
                self._store_nodes(value)

    @override
    @Protect.final
    def on_exiting(self) -> None:
        """Ensure that any unstored nodes in the context are stored, before the state is exited

        After the state is exited the next state will be entered and if persistence is enabled, a checkpoint will
        be saved. If the context contains unstored nodes, the serialization necessary for checkpointing will fail.
        """
        super().on_exiting()
        try:
            self._store_nodes(self.ctx)
        except Exception:  # pylint: disable=broad-except
            # An uncaught exception here will have bizarre and disastrous consequences
            self.logger.exception("exception in _store_nodes called in on_exiting")

    @Protect.final
    def on_wait(self, awaitables: t.Sequence[t.Awaitable]):
        """Entering the WAITING state."""
        super().on_wait(awaitables)
        if self._awaitables:
            self._action_awaitables()
            self.report("Process status: {}".format(self.node.process_status))
        else:
            self.call_soon(self.resume)

    def _action_awaitables(self) -> None:
        """Handle the awaitables that are currently registered with the work chain.

        Depending on the class type of the awaitable's target a different callback
        function will be bound with the awaitable and the runner will be asked to
        call it when the target is completed
        """
        for awaitable in self._awaitables:
            # if the waitable already has a callback, skip
            if awaitable.pk in self.ctx._awaitable_actions:
                continue
            if awaitable.target == AwaitableTarget.PROCESS:
                callback = functools.partial(
                    self.call_soon, self._on_awaitable_finished, awaitable
                )
                self.runner.call_on_process_finish(awaitable.pk, callback)
                self.ctx._awaitable_actions.append(awaitable.pk)
            elif awaitable.target == "asyncio.tasks.Task":
                # this is a awaitable task, the callback function is already set
                self.ctx._awaitable_actions.append(awaitable.pk)
            else:
                assert f"invalid awaitable target '{awaitable.target}'"

    def _on_awaitable_finished(self, awaitable: Awaitable) -> None:
        """Callback function, for when an awaitable process instance is completed.

        The awaitable will be effectuated on the context of the work chain and removed from the internal list. If all
        awaitables have been dealt with, the work chain process is resumed.

        :param awaitable: an Awaitable instance
        """
        self.logger.debug(f"Awaitable {awaitable.key} finished.")

        if isinstance(awaitable.pk, int):
            self.logger.info(
                "received callback that awaitable with key {} and pk {} has terminated".format(
                    awaitable.key, awaitable.pk
                )
            )
            try:
                node = load_node(awaitable.pk)
            except (exceptions.MultipleObjectsError, exceptions.NotExistent):
                raise ValueError(
                    f"provided pk<{awaitable.pk}> could not be resolved to a valid Node instance"
                )

            if awaitable.outputs:
                value = {
                    entry.link_label: entry.node
                    for entry in node.base.links.get_outgoing()
                }
            else:
                value = node  # type: ignore
        else:
            # In this case, the pk and key are the same.
            self.logger.info(
                "received callback that awaitable {} has terminated".format(
                    awaitable.key
                )
            )
            try:
                # if awaitable is cancelled, the result is None
                if awaitable.cancelled():
                    self.set_task_state_info(awaitable.key, "state", "KILLED")
                    # set child tasks state to SKIPPED
                    self.set_tasks_state(
                        self.ctx._connectivity["child_node"][awaitable.key], "SKIPPED"
                    )
                    self.report(f"Task: {awaitable.key} cancelled.")
                else:
                    results = awaitable.result()
                    self.update_normal_task_state(awaitable.key, results)
            except Exception as e:
                self.logger.error(f"Error in awaitable {awaitable.key}: {e}")
                self.set_task_state_info(awaitable.key, "state", "FAILED")
                # set child tasks state to SKIPPED
                self.set_tasks_state(
                    self.ctx._connectivity["child_node"][awaitable.key], "SKIPPED"
                )
                self.report(f"Task: {awaitable.key} failed.")
                self.run_error_handlers(awaitable.key)
            value = None

        self._resolve_awaitable(awaitable, value)

        # node finished, update the task state and result
        # udpate the task state
        self.update_task_state(awaitable.key)
        # try to resume the workgraph, if the workgraph is already resumed
        # by other awaitable, this will not work
        try:
            self.resume()
        except Exception as e:
            print(e)

    def _build_process_label(self) -> str:
        """Use the workgraph name as the process label."""
        return f"WorkGraph<{self.inputs.wg['name']}>"

    def on_create(self) -> None:
        """Called when a Process is created."""
        from aiida_workgraph.utils.analysis import WorkGraphSaver

        super().on_create()
        wgdata = self.inputs.wg._dict
        restart_process = (
            orm.load_node(wgdata["restart_process"].value)
            if wgdata.get("restart_process")
            else None
        )
        saver = WorkGraphSaver(self.node, wgdata, restart_process=restart_process)
        saver.save()
        self.node.label = wgdata["name"]

    def setup(self) -> None:
        """Setup the variables in the context."""
        # track if the awaitable callback is added to the runner
        self.ctx._awaitable_actions = []
        self.ctx._new_data = {}
        self.ctx._executed_tasks = []
        # read the latest workgraph data
        wgdata = self.read_wgdata_from_base()
        self.init_ctx(wgdata)
        #
        self.ctx._msgs = []
        self.ctx._execution_count = 1
        # init task results
        self.set_task_results()
        # data not to be persisted, because they are not serializable
        self._temp = {"awaitables": {}}
        # while workgraph
        if self.ctx._workgraph["workgraph_type"].upper() == "WHILE":
            self.ctx._max_iteration = self.ctx._workgraph.get("max_iteration", 1000)
            should_run = self.check_while_conditions()
            if not should_run:
                self.set_tasks_state(self.ctx._tasks.keys(), "SKIPPED")
        # for workgraph
        if self.ctx._workgraph["workgraph_type"].upper() == "FOR":
            should_run = self.check_for_conditions()
            if not should_run:
                self.set_tasks_state(self.ctx._tasks.keys(), "SKIPPED")

    def setup_ctx_workgraph(self, wgdata: t.Dict[str, t.Any]) -> None:
        """setup the workgraph in the context."""

        self.ctx._tasks = wgdata["tasks"]
        self.ctx._links = wgdata["links"]
        self.ctx._connectivity = wgdata["connectivity"]
        self.ctx._ctrl_links = wgdata["ctrl_links"]
        self.ctx._workgraph = wgdata
        self.ctx._error_handlers = wgdata["error_handlers"]

    def read_wgdata_from_base(self) -> t.Dict[str, t.Any]:
        """Read workgraph data from base.extras."""
        from aiida_workgraph.orm.function_data import PickledLocalFunction

        wgdata = self.node.base.extras.get("_workgraph")
        for name, task in wgdata["tasks"].items():
            wgdata["tasks"][name] = deserialize_unsafe(task)
            for _, input in wgdata["tasks"][name]["inputs"].items():
                if input["property"] is None:
                    continue
                prop = input["property"]
                if isinstance(prop["value"], PickledLocalFunction):
                    prop["value"] = prop["value"].value
        wgdata["error_handlers"] = deserialize_unsafe(wgdata["error_handlers"])
        wgdata["context"] = deserialize_unsafe(wgdata["context"])
        return wgdata

    def update_workgraph_from_base(self) -> None:
        """Update the ctx from base.extras."""
        wgdata = self.read_wgdata_from_base()
        for name, task in wgdata["tasks"].items():
            task["results"] = self.ctx._tasks[name].get("results")
        self.setup_ctx_workgraph(wgdata)

    def get_task(self, name: str):
        """Get task from the context."""
        task = Task.from_dict(self.ctx._tasks[name])
        # update task results
        for output in task.outputs:
            output.value = get_nested_dict(
                self.ctx._tasks[name]["results"],
                output.name,
                default=output.value,
            )
        return task

    def update_task(self, task: Task):
        """Update task in the context.
        This is used in error handlers to update the task parameters."""
        tdata = task.to_dict()
        self.ctx._tasks[task.name]["properties"] = tdata["properties"]
        self.ctx._tasks[task.name]["inputs"] = tdata["inputs"]
        self.reset_task(task.name)

    def get_task_state_info(self, name: str, key: str) -> str:
        """Get task state info from ctx."""

        value = self.ctx._tasks[name].get(key, None)
        if key == "process" and value is not None:
            value = deserialize_unsafe(value)
        return value

    def set_task_state_info(self, name: str, key: str, value: any) -> None:
        """Set task state info to ctx and base.extras.
        We task state to the base.extras, so that we can access outside the engine"""

        if key == "process":
            value = serialize(value)
            self.node.base.extras.set(f"_task_{key}_{name}", value)
        else:
            self.node.base.extras.set(f"_task_{key}_{name}", value)
        self.ctx._tasks[name][key] = value

    def init_ctx(self, wgdata: t.Dict[str, t.Any]) -> None:
        """Init the context from the workgraph data."""
        from aiida_workgraph.utils import update_nested_dict

        # set up the context variables
        self.ctx._max_number_awaitables = (
            wgdata["max_number_jobs"] if wgdata["max_number_jobs"] else 1000000
        )
        self.ctx["_count"] = 0
        for key, value in wgdata["context"].items():
            key = key.replace("__", ".")
            update_nested_dict(self.ctx, key, value)
        # set up the workgraph
        self.setup_ctx_workgraph(wgdata)

    def set_task_results(self) -> None:
        for name, task in self.ctx._tasks.items():
            if self.get_task_state_info(name, "action").upper() == "RESET":
                self.reset_task(task["name"])
            self.update_task_state(name)

    def apply_action(self, msg: dict) -> None:

        if msg["catalog"] == "task":
            self.apply_task_actions(msg)
        else:
            self.report(f"Unknow message type {msg}")

    def apply_task_actions(self, msg: dict) -> None:
        """Apply task actions to the workgraph."""
        action = msg["action"]
        tasks = msg["tasks"]
        self.report(f"Action: {action}. {tasks}")
        if action.upper() == "RESET":
            for name in tasks:
                self.reset_task(name)
        elif action.upper() == "PAUSE":
            for name in tasks:
                self.pause_task(name)
        elif action.upper() == "PLAY":
            for name in tasks:
                self.play_task(name)
        elif action.upper() == "SKIP":
            for name in tasks:
                self.skip_task(name)
        elif action.upper() == "KILL":
            for name in tasks:
                self.kill_task(name)

    def reset_task(
        self,
        name: str,
        reset_process: bool = True,
        recursive: bool = True,
        reset_execution_count: bool = True,
    ) -> None:
        """Reset task state and remove it from the executed task.
        If recursive is True, reset its child tasks."""

        self.set_task_state_info(name, "state", "PLANNED")
        if reset_process:
            self.set_task_state_info(name, "process", None)
        self.remove_executed_task(name)
        # self.logger.debug(f"Task {name} action: RESET.")
        # if the task is a while task, reset its child tasks
        if self.ctx._tasks[name]["metadata"]["node_type"].upper() == "WHILE":
            if reset_execution_count:
                self.ctx._tasks[name]["execution_count"] = 0
            for child_task in self.ctx._tasks[name]["children"]:
                self.reset_task(child_task, reset_process=False, recursive=False)
        elif self.ctx._tasks[name]["metadata"]["node_type"].upper() in ["IF", "ZONE"]:
            for child_task in self.ctx._tasks[name]["children"]:
                self.reset_task(child_task, reset_process=False, recursive=False)
        if recursive:
            # reset its child tasks
            names = self.ctx._connectivity["child_node"][name]
            for name in names:
                self.reset_task(name, recursive=False)

    def pause_task(self, name: str) -> None:
        """Pause task."""
        self.set_task_state_info(name, "action", "PAUSE")
        self.report(f"Task {name} action: PAUSE.")

    def play_task(self, name: str) -> None:
        """Play task."""
        self.set_task_state_info(name, "action", "")
        self.report(f"Task {name} action: PLAY.")

    def skip_task(self, name: str) -> None:
        """Skip task."""
        self.set_task_state_info(name, "state", "SKIPPED")
        self.report(f"Task {name} action: SKIP.")

    def kill_task(self, name: str) -> None:
        """Kill task.
        This is used to kill the awaitable and monitor task.
        """
        if self.get_task_state_info(name, "state") in ["RUNNING"]:
            if self.ctx._tasks[name]["metadata"]["node_type"].upper() in [
                "AWAITABLE",
                "MONITOR",
            ]:
                try:
                    self._temp["awaitables"][name].cancel()
                    self.set_task_state_info(name, "state", "KILLED")
                    self.report(f"Task {name} action: KILLED.")
                except Exception as e:
                    self.logger.error(f"Error in killing task {name}: {e}")

    def continue_workgraph(self) -> None:
        self.report("Continue workgraph.")
        # self.update_workgraph_from_base()
        task_to_run = []
        for name, task in self.ctx._tasks.items():
            # update task state
            if (
                self.get_task_state_info(task["name"], "state")
                in [
                    "CREATED",
                    "RUNNING",
                    "FINISHED",
                    "FAILED",
                    "SKIPPED",
                ]
                or name in self.ctx._executed_tasks
            ):
                continue
            ready, _ = self.is_task_ready_to_run(name)
            if ready:
                task_to_run.append(name)
        #
        self.report("tasks ready to run: {}".format(",".join(task_to_run)))
        self.run_tasks(task_to_run)

    def update_task_state(self, name: str, success=True) -> None:
        """Update task state when the task is finished."""
        task = self.ctx._tasks[name]
        if success:
            node = self.get_task_state_info(name, "process")
            if isinstance(node, orm.ProcessNode):
                # print(f"set task result: {name} process")
                state = node.process_state.value.upper()
                if node.is_finished_ok:
                    self.set_task_state_info(task["name"], "state", state)
                    if task["metadata"]["node_type"].upper() == "WORKGRAPH":
                        # expose the outputs of all the tasks in the workgraph
                        task["results"] = {}
                        outgoing = node.base.links.get_outgoing()
                        for link in outgoing.all():
                            if isinstance(link.node, ProcessNode) and getattr(
                                link.node, "process_state", False
                            ):
                                task["results"][link.link_label] = link.node.outputs
                    else:
                        task["results"] = node.outputs
                        # self.ctx._new_data[name] = task["results"]
                    self.set_task_state_info(task["name"], "state", "FINISHED")
                    self.task_set_context(name)
                    self.report(f"Task: {name} finished.")
                # all other states are considered as failed
                else:
                    task["results"] = node.outputs
                    self.on_task_failed(name)
            elif isinstance(node, orm.Data):
                #
                output_name = [
                    output_name
                    for output_name in list(task["outputs"].keys())
                    if output_name not in ["_wait", "_outputs"]
                ][0]
                task["results"] = {output_name: node}
                self.set_task_state_info(task["name"], "state", "FINISHED")
                self.task_set_context(name)
                self.report(f"Task: {name} finished.")
            else:
                task.setdefault("results", None)
        else:
            self.on_task_failed(name)
        self.update_parent_task_state(name)

    def on_task_failed(self, name: str) -> None:
        """Handle the case where a task has failed."""
        self.set_task_state_info(name, "state", "FAILED")
        self.set_tasks_state(self.ctx._connectivity["child_node"][name], "SKIPPED")
        self.report(f"Task: {name} failed.")
        self.run_error_handlers(name)

    def update_normal_task_state(self, name, results, success=True):
        """Set the results of a normal task.
        A normal task is created by decorating a function with @task().
        """
        from aiida_workgraph.utils import get_sorted_names

        if success:
            task = self.ctx._tasks[name]
            if isinstance(results, tuple):
                if len(task["outputs"]) != len(results):
                    return self.exit_codes.OUTPUS_NOT_MATCH_RESULTS
                output_names = get_sorted_names(task["outputs"])
                for i, output_name in enumerate(output_names):
                    task["results"][output_name] = results[i]
            elif isinstance(results, dict):
                task["results"] = results
            else:
                output_name = [
                    output_name
                    for output_name in list(task["outputs"].keys())
                    if output_name not in ["_wait", "_outputs"]
                ][0]
                task["results"][output_name] = results
            self.task_set_context(name)
            self.set_task_state_info(name, "state", "FINISHED")
            self.report(f"Task: {name} finished.")
        else:
            self.on_task_failed(name)
        self.update_parent_task_state(name)

    def update_parent_task_state(self, name: str) -> None:
        """Update parent task state."""
        parent_task = self.ctx._tasks[name]["parent_task"]
        if parent_task[0]:
            task_type = self.ctx._tasks[parent_task[0]]["metadata"]["node_type"].upper()
            if task_type == "WHILE":
                self.update_while_task_state(parent_task[0])
            elif task_type == "IF":
                self.update_zone_task_state(parent_task[0])
            elif task_type == "ZONE":
                self.update_zone_task_state(parent_task[0])

    def update_while_task_state(self, name: str) -> None:
        """Update while task state."""
        finished, _ = self.are_childen_finished(name)

        if finished:
            self.report(
                f"Wihle Task {name}: this iteration finished. Try to reset for the next iteration."
            )
            # reset the condition tasks
            for link in self.ctx._tasks[name]["inputs"]["conditions"]["links"]:
                self.reset_task(link["from_node"], recursive=False)
            # reset the task and all its children, so that the task can run again
            # do not reset the execution count
            self.reset_task(name, reset_execution_count=False)

    def update_zone_task_state(self, name: str) -> None:
        """Update zone task state."""
        finished, _ = self.are_childen_finished(name)
        if finished:
            self.set_task_state_info(name, "state", "FINISHED")
            self.report(f"Task: {name} finished.")
            self.update_parent_task_state(name)

    def should_run_while_task(self, name: str) -> tuple[bool, t.Any]:
        """Check if the while task should run."""
        # check the conditions of the while task
        not_excess_max_iterations = (
            self.ctx._tasks[name]["execution_count"]
            < self.ctx._tasks[name]["inputs"]["max_iterations"]["property"]["value"]
        )
        conditions = [not_excess_max_iterations]
        _, kwargs, _, _, _ = self.get_inputs(name)
        if isinstance(kwargs["conditions"], list):
            for condition in kwargs["conditions"]:
                value = get_nested_dict(self.ctx, condition)
                conditions.append(value)
        elif isinstance(kwargs["conditions"], dict):
            for _, value in kwargs["conditions"].items():
                conditions.append(value)
        else:
            conditions.append(kwargs["conditions"])
        return False not in conditions

    def should_run_if_task(self, name: str) -> tuple[bool, t.Any]:
        """Check if the IF task should run."""
        _, kwargs, _, _, _ = self.get_inputs(name)
        flag = kwargs["conditions"]
        if kwargs["invert_condition"]:
            return not flag
        return flag

    def are_childen_finished(self, name: str) -> tuple[bool, t.Any]:
        """Check if the child tasks are finished."""
        task = self.ctx._tasks[name]
        finished = True
        for name in task["children"]:
            if self.get_task_state_info(name, "state") not in [
                "FINISHED",
                "SKIPPED",
                "FAILED",
            ]:
                finished = False
                break
        return finished, None

    def run_error_handlers(self, task_name: str) -> None:
        """Run error handler for a task."""

        node = self.get_task_state_info(task_name, "process")
        if not node or not node.exit_status:
            return
        # error_handlers from the task
        for _, data in self.ctx._tasks[task_name]["error_handlers"].items():
            if node.exit_status in data.get("exit_codes", []):
                handler = data["handler"]
                self.run_error_handler(handler, data, task_name)
                return
        # error_handlers from the workgraph
        for _, data in self.ctx._error_handlers.items():
            if node.exit_code.status in data["tasks"].get(task_name, {}).get(
                "exit_codes", []
            ):
                handler = data["handler"]
                metadata = data["tasks"][task_name]
                self.run_error_handler(handler, metadata, task_name)
                return

    def run_error_handler(self, handler: dict, metadata: dict, task_name: str) -> None:
        from inspect import signature
        from aiida_workgraph.utils import get_executor

        handler, _ = get_executor(handler)
        handler_sig = signature(handler)
        metadata.setdefault("retry", 0)
        self.report(f"Run error handler: {handler.__name__}")
        if metadata["retry"] < metadata["max_retries"]:
            task = self.get_task(task_name)
            try:
                if "engine" in handler_sig.parameters:
                    msg = handler(task, engine=self, **metadata.get("kwargs", {}))
                else:
                    msg = handler(task, **metadata.get("kwargs", {}))
                self.update_task(task)
                if msg:
                    self.report(msg)
                metadata["retry"] += 1
            except Exception as e:
                self.report(f"Error in running error handler: {e}")

    def is_workgraph_finished(self) -> bool:
        """Check if the workgraph is finished.
        For `while` workgraph, we need check its conditions"""
        is_finished = True
        failed_tasks = []
        for name, task in self.ctx._tasks.items():
            # self.update_task_state(name)
            if self.get_task_state_info(task["name"], "state") in [
                "RUNNING",
                "CREATED",
                "PLANNED",
                "READY",
            ]:
                is_finished = False
            elif self.get_task_state_info(task["name"], "state") == "FAILED":
                failed_tasks.append(name)
        if is_finished:
            if self.ctx._workgraph["workgraph_type"].upper() == "WHILE":
                should_run = self.check_while_conditions()
                is_finished = not should_run
            if self.ctx._workgraph["workgraph_type"].upper() == "FOR":
                should_run = self.check_for_conditions()
                is_finished = not should_run
        if is_finished and len(failed_tasks) > 0:
            message = f"WorkGraph finished, but tasks: {failed_tasks} failed. Thus all their child tasks are skipped."
            self.report(message)
            result = ExitCode(302, message)
        else:
            result = None
        return is_finished, result

    def check_while_conditions(self) -> bool:
        """Check while conditions.
        Run all condition tasks and check if all the conditions are True.
        """
        self.report("Check while conditions.")
        if self.ctx._execution_count >= self.ctx._max_iteration:
            self.report("Max iteration reached.")
            return False
        condition_tasks = []
        for c in self.ctx._workgraph["conditions"]:
            task_name, socket_name = c.split(".")
            if "task_name" != "context":
                condition_tasks.append(task_name)
                self.reset_task(task_name)
        self.run_tasks(condition_tasks, continue_workgraph=False)
        conditions = []
        for c in self.ctx._workgraph["conditions"]:
            task_name, socket_name = c.split(".")
            if task_name == "context":
                conditions.append(self.ctx[socket_name])
            else:
                conditions.append(self.ctx._tasks[task_name]["results"][socket_name])
        should_run = False not in conditions
        if should_run:
            self.reset()
            self.set_tasks_state(condition_tasks, "SKIPPED")
        return should_run

    def check_for_conditions(self) -> bool:
        condition_tasks = [c[0] for c in self.ctx._workgraph["conditions"]]
        self.run_tasks(condition_tasks)
        conditions = [self.ctx._count < len(self.ctx._sequence)] + [
            self.ctx._tasks[c[0]]["results"][c[1]]
            for c in self.ctx._workgraph["conditions"]
        ]
        should_run = False not in conditions
        if should_run:
            self.reset()
            self.set_tasks_state(condition_tasks, "SKIPPED")
            self.ctx["i"] = self.ctx._sequence[self.ctx._count]
        self.ctx._count += 1
        return should_run

    def remove_executed_task(self, name: str) -> None:
        """Remove labels with name from executed tasks."""
        self.ctx._executed_tasks = [
            label for label in self.ctx._executed_tasks if label.split(".")[0] != name
        ]

    def run_tasks(self, names: t.List[str], continue_workgraph: bool = True) -> None:
        """Run tasks.
        Task type includes: Node, Data, CalcFunction, WorkFunction, CalcJob, WorkChain, GraphBuilder,
        WorkGraph, PythonJob, ShellJob, While, If, Zone, FromContext, ToContext, Normal.

        Here we use ToContext to pass the results of the run to the next step.
        This will force the engine to wait for all the submitted processes to
        finish before continuing to the next step.
        """
        from aiida_workgraph.utils import (
            get_executor,
            create_data_node,
            update_nested_dict_with_special_keys,
        )

        for name in names:
            # skip if the max number of awaitables is reached
            task = self.ctx._tasks[name]
            if task["metadata"]["node_type"].upper() in [
                "CALCJOB",
                "WORKCHAIN",
                "GRAPH_BUILDER",
                "WORKGRAPH",
                "PYTHONJOB",
                "SHELLJOB",
            ]:
                if len(self._awaitables) >= self.ctx._max_number_awaitables:
                    print(
                        MAX_NUMBER_AWAITABLES_MSG.format(
                            self.ctx._max_number_awaitables, name
                        )
                    )
                    continue
            # skip if the task is already executed
            # or if the task is in a skippped state
            if name in self.ctx._executed_tasks or self.get_task_state_info(
                name, "state"
            ) in ["SKIPPED"]:
                continue
            self.ctx._executed_tasks.append(name)
            print("-" * 60)

            self.report(f"Run task: {name}, type: {task['metadata']['node_type']}")
            executor, _ = get_executor(task["executor"])
            # print("executor: ", executor)
            args, kwargs, var_args, var_kwargs, args_dict = self.get_inputs(name)
            for i, key in enumerate(self.ctx._tasks[name]["args"]):
                kwargs[key] = args[i]
            # update the port namespace
            kwargs = update_nested_dict_with_special_keys(kwargs)
            # print("args: ", args)
            # print("kwargs: ", kwargs)
            # print("var_kwargs: ", var_kwargs)
            # kwargs["meta.label"] = name
            # output must be a Data type or a mapping of {string: Data}
            task["results"] = {}
            if task["metadata"]["node_type"].upper() == "NODE":
                results = self.run_executor(executor, [], kwargs, var_args, var_kwargs)
                self.set_task_state_info(name, "process", results)
                self.update_task_state(name)
                if continue_workgraph:
                    self.continue_workgraph()
            elif task["metadata"]["node_type"].upper() == "DATA":
                for key in self.ctx._tasks[name]["args"]:
                    kwargs.pop(key, None)
                results = create_data_node(executor, args, kwargs)
                self.set_task_state_info(name, "process", results)
                self.update_task_state(name)
                self.ctx._new_data[name] = results
                if continue_workgraph:
                    self.continue_workgraph()
            elif task["metadata"]["node_type"].upper() in [
                "CALCFUNCTION",
                "WORKFUNCTION",
            ]:
                kwargs.setdefault("metadata", {})
                kwargs["metadata"].update({"call_link_label": name})
                try:
                    # since aiida 2.5.0, we need to use args_dict to pass the args to the run_get_node
                    if var_kwargs is None:
                        results, process = run_get_node(executor, **kwargs)
                    else:
                        results, process = run_get_node(
                            executor, **kwargs, **var_kwargs
                        )
                    process.label = name
                    # print("results: ", results)
                    self.set_task_state_info(name, "process", process)
                    self.update_task_state(name)
                except Exception as e:
                    self.logger.error(f"Error in task {name}: {e}")
                    self.update_task_state(name, success=False)
                # exclude the current tasks from the next run
                if continue_workgraph:
                    self.continue_workgraph()
            elif task["metadata"]["node_type"].upper() in ["CALCJOB", "WORKCHAIN"]:
                # process = run_get_node(executor, *args, **kwargs)
                kwargs.setdefault("metadata", {})
                kwargs["metadata"].update({"call_link_label": name})
                # transfer the args to kwargs
                if self.get_task_state_info(name, "action").upper() == "PAUSE":
                    self.set_task_state_info(name, "action", "")
                    self.report(f"Task {name} is created and paused.")
                    process = create_and_pause_process(
                        self.runner,
                        executor,
                        kwargs,
                        state_msg="Paused through WorkGraph",
                    )
                    self.set_task_state_info(name, "state", "CREATED")
                    process = process.node
                else:
                    process = self.submit(executor, **kwargs)
                    self.set_task_state_info(name, "state", "RUNNING")
                process.label = name
                self.set_task_state_info(name, "process", process)
                self.to_context(**{name: process})
            elif task["metadata"]["node_type"].upper() in ["GRAPH_BUILDER"]:
                wg = self.run_executor(executor, [], kwargs, var_args, var_kwargs)
                wg.name = name
                wg.group_outputs = self.ctx._tasks[name]["metadata"]["group_outputs"]
                wg.parent_uuid = self.node.uuid
                inputs = wg.prepare_inputs(metadata={"call_link_label": name})
                process = self.submit(WorkGraphEngine, inputs=inputs)
                self.set_task_state_info(name, "process", process)
                self.set_task_state_info(name, "state", "RUNNING")
                self.to_context(**{name: process})
            elif task["metadata"]["node_type"].upper() in ["WORKGRAPH"]:
                from .utils import prepare_for_workgraph_task

                inputs, _ = prepare_for_workgraph_task(task, kwargs)
                process = self.submit(WorkGraphEngine, inputs=inputs)
                self.set_task_state_info(name, "process", process)
                self.set_task_state_info(name, "state", "RUNNING")
                self.to_context(**{name: process})
            elif task["metadata"]["node_type"].upper() in ["PYTHONJOB"]:
                from aiida_workgraph.calculations.python import PythonJob
                from .utils import prepare_for_python_task

                inputs = prepare_for_python_task(task, kwargs, var_kwargs)
                # since aiida 2.5.0, we can pass inputs directly to the submit, no need to use **inputs
                if self.get_task_state_info(name, "action").upper() == "PAUSE":
                    self.set_task_state_info(name, "action", "")
                    self.report(f"Task {name} is created and paused.")
                    process = create_and_pause_process(
                        self.runner,
                        PythonJob,
                        inputs,
                        state_msg="Paused through WorkGraph",
                    )
                    self.set_task_state_info(name, "state", "CREATED")
                    process = process.node
                else:
                    process = self.submit(PythonJob, **inputs)
                    self.set_task_state_info(name, "state", "RUNNING")
                process.label = name
                self.set_task_state_info(name, "process", process)
                self.to_context(**{name: process})
            elif task["metadata"]["node_type"].upper() in ["SHELLJOB"]:
                from aiida_shell.calculations.shell import ShellJob
                from .utils import prepare_for_shell_task

                inputs = prepare_for_shell_task(task, kwargs)
                if self.get_task_state_info(name, "action").upper() == "PAUSE":
                    self.set_task_state_info(name, "action", "")
                    self.report(f"Task {name} is created and paused.")
                    process = create_and_pause_process(
                        self.runner,
                        ShellJob,
                        inputs,
                        state_msg="Paused through WorkGraph",
                    )
                    self.set_task_state_info(name, "state", "CREATED")
                    process = process.node
                else:
                    process = self.submit(ShellJob, **inputs)
                    self.set_task_state_info(name, "state", "RUNNING")
                process.label = name
                self.set_task_state_info(name, "process", process)
                self.to_context(**{name: process})
            elif task["metadata"]["node_type"].upper() in ["WHILE"]:
                # TODO refactor this for while, if and zone
                # in case of an empty zone, it will finish immediately
                if self.are_childen_finished(name)[0]:
                    self.update_while_task_state(name)
                else:
                    # check the conditions of the while task
                    should_run = self.should_run_while_task(name)
                    if not should_run:
                        self.set_task_state_info(name, "state", "FINISHED")
                        self.set_tasks_state(
                            self.ctx._tasks[name]["children"], "SKIPPED"
                        )
                        self.update_parent_task_state(name)
                        self.report(
                            f"While Task {name}: Condition not fullilled, task finished. Skip all its children."
                        )
                    else:
                        task["execution_count"] += 1
                        self.set_task_state_info(name, "state", "RUNNING")
                self.continue_workgraph()
            elif task["metadata"]["node_type"].upper() in ["IF"]:
                # in case of an empty zone, it will finish immediately
                if self.are_childen_finished(name)[0]:
                    self.update_zone_task_state(name)
                else:
                    should_run = self.should_run_if_task(name)
                    if should_run:
                        self.set_task_state_info(name, "state", "RUNNING")
                    else:
                        self.set_tasks_state(task["children"], "SKIPPED")
                        self.update_zone_task_state(name)
                self.continue_workgraph()
            elif task["metadata"]["node_type"].upper() in ["ZONE"]:
                # in case of an empty zone, it will finish immediately
                if self.are_childen_finished(name)[0]:
                    self.update_zone_task_state(name)
                else:
                    self.set_task_state_info(name, "state", "RUNNING")
                self.continue_workgraph()
            elif task["metadata"]["node_type"].upper() in ["FROM_CONTEXT"]:
                # get the results from the context
                results = {"result": getattr(self.ctx, kwargs["key"])}
                task["results"] = results
                self.set_task_state_info(name, "state", "FINISHED")
                self.update_parent_task_state(name)
                self.continue_workgraph()
            elif task["metadata"]["node_type"].upper() in ["TO_CONTEXT"]:
                # get the results from the context
                setattr(self.ctx, kwargs["key"], kwargs["value"])
                self.set_task_state_info(name, "state", "FINISHED")
                self.update_parent_task_state(name)
                self.continue_workgraph()
            elif task["metadata"]["node_type"].upper() in ["AWAITABLE"]:
                for key in self.ctx._tasks[name]["args"]:
                    kwargs.pop(key, None)
                awaitable_target = asyncio.ensure_future(
                    self.run_executor(executor, args, kwargs, var_args, var_kwargs),
                    loop=self.loop,
                )
                awaitable = self.construct_awaitable_function(name, awaitable_target)
                self.set_task_state_info(name, "state", "RUNNING")
                self.to_context(**{name: awaitable})
            elif task["metadata"]["node_type"].upper() in ["MONITOR"]:

                for key in self.ctx._tasks[name]["args"]:
                    kwargs.pop(key, None)
                # add function and interval to the args
                args = [
                    executor,
                    kwargs.pop("interval", 1),
                    kwargs.pop("timeout", 3600),
                    *args,
                ]
                awaitable_target = asyncio.ensure_future(
                    self.run_executor(monitor, args, kwargs, var_args, var_kwargs),
                    loop=self.loop,
                )
                awaitable = self.construct_awaitable_function(name, awaitable_target)
                self.set_task_state_info(name, "state", "RUNNING")
                # save the awaitable to the temp, so that we can kill it if needed
                self._temp["awaitables"][name] = awaitable_target
                self.to_context(**{name: awaitable})
            elif task["metadata"]["node_type"].upper() in ["NORMAL"]:
                # Normal task is created by decoratoring a function with @task()
                if "context" in task["kwargs"]:
                    self.ctx.task_name = name
                    kwargs.update({"context": self.ctx})
                for key in self.ctx._tasks[name]["args"]:
                    kwargs.pop(key, None)
                try:
                    results = self.run_executor(
                        executor, args, kwargs, var_args, var_kwargs
                    )
                    self.update_normal_task_state(name, results)
                except Exception as e:
                    self.logger.error(f"Error in task {name}: {e}")
                    self.update_normal_task_state(name, results=None, success=False)
                if continue_workgraph:
                    self.continue_workgraph()
            else:
                # self.report("Unknow task type {}".format(task["metadata"]["node_type"]))
                return self.exit_codes.UNKNOWN_TASK_TYPE

    def construct_awaitable_function(
        self, name: str, awaitable_target: Awaitable
    ) -> None:
        """Construct the awaitable function."""
        awaitable = Awaitable(
            **{
                "pk": name,
                "action": AwaitableAction.ASSIGN,
                "target": "asyncio.tasks.Task",
                "outputs": False,
            }
        )
        awaitable_target.key = name
        awaitable_target.pk = name
        awaitable_target.action = AwaitableAction.ASSIGN
        awaitable_target.add_done_callback(self._on_awaitable_finished)
        return awaitable

    def get_inputs(
        self, name: str
    ) -> t.Tuple[
        t.List[t.Any],
        t.Dict[str, t.Any],
        t.Optional[t.List[t.Any]],
        t.Optional[t.Dict[str, t.Any]],
        t.Dict[str, t.Any],
    ]:
        """Get input based on the links."""

        args = []
        args_dict = {}
        kwargs = {}
        var_args = None
        var_kwargs = None
        task = self.ctx._tasks[name]
        properties = task.get("properties", {})
        inputs = {}
        for name, input in task["inputs"].items():
            # print(f"input: {input['name']}")
            if len(input["links"]) == 0:
                inputs[name] = self.update_context_variable(input["property"]["value"])
            elif len(input["links"]) == 1:
                link = input["links"][0]
                if self.ctx._tasks[link["from_node"]]["results"] is None:
                    inputs[name] = None
                else:
                    # handle the special socket _wait, _outputs
                    if link["from_socket"] == "_wait":
                        continue
                    elif link["from_socket"] == "_outputs":
                        inputs[name] = self.ctx._tasks[link["from_node"]]["results"]
                    else:
                        inputs[name] = get_nested_dict(
                            self.ctx._tasks[link["from_node"]]["results"],
                            link["from_socket"],
                        )
            # handle the case of multiple outputs
            elif len(input["links"]) > 1:
                value = {}
                for link in input["links"]:
                    item_name = f'{link["from_node"]}_{link["from_socket"]}'
                    # handle the special socket _wait, _outputs
                    if link["from_socket"] == "_wait":
                        continue
                    if self.ctx._tasks[link["from_node"]]["results"] is None:
                        value[item_name] = None
                    else:
                        value[item_name] = self.ctx._tasks[link["from_node"]][
                            "results"
                        ][link["from_socket"]]
                inputs[name] = value
        for name in task.get("args", []):
            if name in inputs:
                args.append(inputs[name])
                args_dict[name] = inputs[name]
            else:
                value = self.update_context_variable(properties[name]["value"])
                args.append(value)
                args_dict[name] = value
        for name in task.get("kwargs", []):
            if name in inputs:
                kwargs[name] = inputs[name]
            else:
                value = self.update_context_variable(properties[name]["value"])
                kwargs[name] = value
        if task["var_args"] is not None:
            name = task["var_args"]
            if name in inputs:
                var_args = inputs[name]
            else:
                value = self.update_context_variable(properties[name]["value"])
                var_args = value
        if task["var_kwargs"] is not None:
            name = task["var_kwargs"]
            if name in inputs:
                var_kwargs = inputs[name]
            else:
                value = self.update_context_variable(properties[name]["value"])
                var_kwargs = value
        return args, kwargs, var_args, var_kwargs, args_dict

    def update_context_variable(self, value: t.Any) -> t.Any:
        # replace context variables

        """Get value from context."""
        if isinstance(value, dict):
            for key, sub_value in value.items():
                value[key] = self.update_context_variable(sub_value)
        elif (
            isinstance(value, str)
            and value.strip().startswith("{{")
            and value.strip().endswith("}}")
        ):
            name = value[2:-2].strip()
            return get_nested_dict(self.ctx, name)
        return value

    def task_set_context(self, name: str) -> None:
        """Export task result to context."""
        from aiida_workgraph.utils import update_nested_dict

        items = self.ctx._tasks[name]["context_mapping"]
        for key, value in items.items():
            result = self.ctx._tasks[name]["results"][key]
            update_nested_dict(self.ctx, value, result)

    def is_task_ready_to_run(self, name: str) -> t.Tuple[bool, t.Optional[str]]:
        """Check if the task ready to run.
        For normal task and a zone task, we need to check its input tasks in the connectivity["zone"].
        For task inside a zone, we need to check if the zone (parent task) is ready.
        """
        parent_task = self.ctx._tasks[name]["parent_task"]
        # input_tasks, parent_task, conditions
        parent_states = [True, True]
        # if the task belongs to a parent zone
        if parent_task[0]:
            state = self.get_task_state_info(parent_task[0], "state")
            if state not in ["RUNNING"]:
                parent_states[1] = False
        # check the input tasks of the zone
        # check if the zone input tasks are ready
        for child_task_name in self.ctx._connectivity["zone"][name]["input_tasks"]:
            if self.get_task_state_info(child_task_name, "state") not in [
                "FINISHED",
                "SKIPPED",
                "FAILED",
            ]:
                parent_states[0] = False
                break

        return all(parent_states), parent_states

    def reset(self) -> None:
        self.ctx._execution_count += 1
        self.set_tasks_state(self.ctx._tasks.keys(), "PLANNED")
        self.ctx._executed_tasks = []

    def set_tasks_state(
        self, tasks: t.Union[t.List[str], t.Sequence[str]], value: str
    ) -> None:
        """Set tasks state"""
        for name in tasks:
            self.set_task_state_info(name, "state", value)
            if "children" in self.ctx._tasks[name]:
                self.set_tasks_state(self.ctx._tasks[name]["children"], value)

    def run_executor(
        self,
        executor: t.Callable,
        args: t.List[t.Any],
        kwargs: t.Dict[str, t.Any],
        var_args: t.Optional[t.List[t.Any]],
        var_kwargs: t.Optional[t.Dict[str, t.Any]],
    ) -> t.Any:
        if var_kwargs is None:
            return executor(*args, **kwargs)
        else:
            return executor(*args, **kwargs, **var_kwargs)

    def save_results_to_extras(self, name: str) -> None:
        """Save the results to the base.extras.
        For the outputs of a Normal task, they are not saved to the database like the calcjob or workchain.
        One temporary solution is to save the results to the base.extras. In order to do this, we need to
        serialize the results
        """
        from aiida_workgraph.utils import get_executor

        results = self.ctx._tasks[name]["results"]
        if results is None:
            return
        datas = {}
        for key, value in results.items():
            # find outptus sockets with the name as key
            output = [
                output
                for output in self.ctx._tasks[name]["outputs"]
                if output["name"] == key
            ]
            if len(output) == 0:
                continue
            output = output[0]
            Executor, _ = get_executor(output["serialize"])
            datas[key] = Executor(value)
        self.node.set_extra(f"nodes__results__{name}", datas)

    def message_receive(
        self, _comm: kiwipy.Communicator, msg: t.Dict[str, t.Any]
    ) -> t.Any:
        """
        Coroutine called when the process receives a message from the communicator

        :param _comm: the communicator that sent the message
        :param msg: the message
        :return: the outcome of processing the message, the return value will be sent back as a response to the sender
        """
        self.logger.debug(
            "Process<%s>: received RPC message with communicator '%s': %r",
            self.pid,
            _comm,
            msg,
        )

        intent = msg[process_comms.INTENT_KEY]

        if intent == process_comms.Intent.PLAY:
            return self._schedule_rpc(self.play)
        if intent == process_comms.Intent.PAUSE:
            return self._schedule_rpc(
                self.pause, msg=msg.get(process_comms.MESSAGE_KEY, None)
            )
        if intent == process_comms.Intent.KILL:
            return self._schedule_rpc(
                self.kill, msg=msg.get(process_comms.MESSAGE_KEY, None)
            )
        if intent == process_comms.Intent.STATUS:
            status_info: t.Dict[str, t.Any] = {}
            self.get_status_info(status_info)
            return status_info
        if intent == "custom":
            return self._schedule_rpc(self.apply_action, msg=msg)

        # Didn't match any known intents
        raise RuntimeError("Unknown intent")

    def finalize(self) -> t.Optional[ExitCode]:
        """"""
        # expose outputs of the workgraph
        group_outputs = {}
        for output in self.ctx._workgraph["metadata"]["group_outputs"]:
            names = output["from"].split(".", 1)
            if names[0] == "context":
                if len(names) == 1:
                    raise ValueError("The output name should be context.key")
                update_nested_dict(
                    group_outputs,
                    output["name"],
                    get_nested_dict(self.ctx, names[1]),
                )
            else:
                # expose the whole outputs of the tasks
                if len(names) == 1:
                    update_nested_dict(
                        group_outputs,
                        output["name"],
                        self.ctx._tasks[names[0]]["results"],
                    )
                else:
                    # expose one output of the task
                    # note, the output may not exist
                    if names[1] in self.ctx._tasks[names[0]]["results"]:
                        update_nested_dict(
                            group_outputs,
                            output["name"],
                            self.ctx._tasks[names[0]]["results"][names[1]],
                        )
        self.out_many(group_outputs)
        # output the new data
        if self.ctx._new_data:
            self.out("new_data", self.ctx._new_data)
        self.out("execution_count", orm.Int(self.ctx._execution_count).store())
        self.report("Finalize workgraph.")
        for _, task in self.ctx._tasks.items():
            if self.get_task_state_info(task["name"], "state") == "FAILED":
                return self.exit_codes.TASK_FAILED



from aiida.manage import get_manager
from aiida import orm
from aiida.engine.processes import control


def create_task_action(
    pk: int,
    tasks: list,
    action: str = "pause",
    timeout: int = 5,
    wait: bool = False,
):
    """Send task action to Process."""

    controller = get_manager().get_process_controller()
    message = {"intent": "custom", "catalog": "task", "action": action, "tasks": tasks}
    controller._communicator.rpc_send(pk, message)


def get_task_state_info(node, name: str, key: str) -> str:
    """Get task state info from base.extras."""
    from aiida.orm.utils.serialize import deserialize_unsafe

    if key == "process":
        value = deserialize_unsafe(node.base.extras.get(f"_task_{key}_{name}", ""))
    else:
        value = node.base.extras.get(f"_task_{key}_{name}", "")
    return value


def set_task_state_info(node, name: str, key: str, value: any) -> None:
    """Set task state info to base.extras."""
    from aiida.orm.utils.serialize import serialize

    if key == "process":
        node.base.extras.set(f"_task_{key}_{name}", serialize(value))
    else:
        node.base.extras.set(f"_task_{key}_{name}", value)


def pause_tasks(pk: int, tasks: list, timeout: int = 5, wait: bool = False):
    """Pause task."""
    node = orm.load_node(pk)
    if node.is_finished:
        message = "Process is finished. Cannot pause tasks."
        print(message)
        return False, message
    elif node.process_state.value.upper() in [
        "CREATED",
        "RUNNING",
        "WAITING",
        "PAUSED",
    ]:
        for name in tasks:
            if get_task_state_info(node, name, "state") == "PLANNED":
                create_task_action(pk, tasks, action="pause")
            elif get_task_state_info(node, name, "state") == "RUNNING":
                try:
                    control.pause_processes(
                        [get_task_state_info(node, name, "process")],
                        all_entries=None,
                        timeout=5,
                        wait=False,
                    )
                except Exception as e:
                    print(f"Pause task {name} failed: {e}")
    return True, ""


def play_tasks(pk: int, tasks: list, timeout: int = 5, wait: bool = False):
    node = orm.load_node(pk)
    if node.is_finished:
        message = "Process is finished. Cannot pause tasks."
        print(message)
        return False, message
    elif node.process_state.value.upper() in [
        "CREATED",
        "RUNNING",
        "WAITING",
        "PAUSED",
    ]:
        for name in tasks:
            if get_task_state_info(node, name, "state") == "PLANNED":
                create_task_action(pk, tasks, action="play")
            elif get_task_state_info(node, name, "state") in ["CREATED", "PAUSED"]:
                try:
                    control.play_processes(
                        [get_task_state_info(node, name, "process")],
                        all_entries=None,
                        timeout=5,
                        wait=False,
                    )
                except Exception as e:
                    print(f"Play task {name} failed: {e}")
            elif get_task_state_info(node, name, "process").is_finished:
                raise ValueError(f"Task {name} is already finished.")
    return True, ""


def kill_tasks(pk: int, tasks: list, timeout: int = 5, wait: bool = False):
    node = orm.load_node(pk)
    if node.is_finished:
        message = "Process is finished. Cannot pause tasks."
        print(message)
        return False, message
    elif node.process_state.value.upper() in [
        "CREATED",
        "RUNNING",
        "WAITING",
        "PAUSED",
    ]:
        print("tasks", tasks)
        for name in tasks:
            state = get_task_state_info(node, name, "state")
            process = get_task_state_info(node, name, "process")
            print("state", state)
            print("process", process)
            if state == "PLANNED":
                create_task_action(pk, tasks, action="skip")
            elif state in [
                "CREATED",
                "RUNNING",
                "WAITING",
                "PAUSED",
            ]:
                if get_task_state_info(node, name, "process") is None:
                    print(f"Task {name} is not a AiiDA process.")
                    create_task_action(pk, tasks, action="kill")
                else:
                    try:
                        control.kill_processes(
                            [get_task_state_info(node, name, "process")],
                            all_entries=None,
                            timeout=5,
                            wait=False,
                        )
                    except Exception as e:
                        print(f"Kill task {name} failed: {e}")
    return True, ""



import json
from aiida.manage.configuration.settings import AIIDA_CONFIG_FOLDER


def load_config() -> dict:
    """Load the configuration from the config file."""
    config_file_path = AIIDA_CONFIG_FOLDER / "workgraph.json"
    try:
        with config_file_path.open("r") as f:
            config = json.load(f)
    except FileNotFoundError:
        config = {}
    return config



"""`verdi process` command."""
import click

from aiida_workgraph.cli.cmd_workgraph import workgraph
from aiida.cmdline.params import arguments, options, types
from aiida.cmdline.utils import decorators, echo
from aiida.common.log import LOG_LEVELS
from aiida.manage import get_manager
from aiida_workgraph.cli.query_workgraph import WorkGraphQueryBuilder

REPAIR_INSTRUCTIONS = """\
If one ore more processes are unreachable, you can run the following commands to try and repair them:

    verdi daemon stop
    verdi process repair
    verdi daemon start
"""


def valid_projections():
    """Return list of valid projections for the ``--project`` option of ``verdi process list``.

    This indirection is necessary to prevent loading the imported module which slows down tab-completion.
    """
    return WorkGraphQueryBuilder.valid_projections


def default_projections():
    """Return list of default projections for the ``--project`` option of ``verdi process list``.

    This indirection is necessary to prevent loading the imported module which slows down tab-completion.
    """
    return WorkGraphQueryBuilder.default_projections


@workgraph.group("graph")
def workgraph_graph():
    """Inspect and manage processes."""


@workgraph_graph.command("list")
@options.PROJECT(
    type=types.LazyChoice(valid_projections), default=lambda: default_projections()
)  # pylint: disable=unnecessary-lambda
@options.ORDER_BY()
@options.ORDER_DIRECTION()
@options.GROUP(help="Only include entries that are a member of this group.")
@options.ALL(help="Show all entries, regardless of their process state.")
@options.PROCESS_STATE()
@options.PROCESS_LABEL()
@options.PAUSED()
@options.EXIT_STATUS()
@options.FAILED()
@options.PAST_DAYS()
@options.LIMIT()
@options.RAW()
@click.pass_context
@decorators.with_dbenv()
def process_list(
    ctx,
    all_entries,
    group,
    process_state,
    process_label,
    paused,
    exit_status,
    failed,
    past_days,
    limit,
    project,
    raw,
    order_by,
    order_dir,
):
    """Show a list of running or terminated WorkGraph processes.

    By default, only those that are still running are shown, but there are options to show also the finished ones.
    """
    # pylint: disable=too-many-locals
    from tabulate import tabulate

    from aiida.cmdline.commands.cmd_daemon import execute_client_command
    from aiida.cmdline.utils.common import print_last_process_state_change
    from aiida.engine.daemon.client import get_daemon_client
    from aiida.orm import ProcessNode, QueryBuilder

    relationships = {}

    if group:
        relationships["with_node"] = group

    builder = WorkGraphQueryBuilder()
    filters = builder.get_filters(
        all_entries, process_state, process_label, paused, exit_status, failed
    )
    query_set = builder.get_query_set(
        relationships=relationships,
        filters=filters,
        order_by={order_by: order_dir},
        past_days=past_days,
        limit=limit,
    )
    projected = builder.get_projected(query_set, projections=project)
    headers = projected.pop(0)

    if raw:
        tabulated = tabulate(projected, tablefmt="plain")
        echo.echo(tabulated)
        return

    tabulated = tabulate(projected, headers=headers)
    echo.echo(tabulated)
    echo.echo(f"\nTotal results: {len(projected)}\n")

    if "cached" in project:
        echo.echo_report(
            "\u267B Processes marked with check-mark were not run but taken from the cache."
        )
        echo.echo_report(
            "Add the option `-P pk cached_from` to the command to display cache source."
        )

    print_last_process_state_change()

    if not get_daemon_client().is_daemon_running:
        echo.echo_warning("The daemon is not running", bold=True)
        return

    echo.echo_report("Checking daemon load... ", nl=False)
    response = execute_client_command("get_numprocesses")

    if not response:
        # Daemon could not be reached
        return

    try:
        active_workers = response["numprocesses"]
    except KeyError:
        echo.echo_report("No active daemon workers.")
    else:
        # Second query to get active process count. Currently this is slow but will be fixed with issue #2770. It is
        # placed at the end of the command so that the user can Ctrl+C after getting the process table.
        slots_per_worker = ctx.obj.config.get_option(
            "daemon.worker_process_slots", scope=ctx.obj.profile.name
        )
        active_processes = (
            QueryBuilder()
            .append(
                ProcessNode,
                filters={
                    "attributes.process_state": {
                        "in": ("created", "waiting", "running")
                    }
                },
            )
            .count()
        )
        available_slots = active_workers * slots_per_worker
        percent_load = active_processes / available_slots
        if percent_load > 0.9:  # 90%
            echo.echo_warning(
                f"{percent_load * 100:.0f}% of the available daemon worker slots have been used!"
            )
            echo.echo_warning(
                "Increase the number of workers with `verdi daemon incr`."
            )
        else:
            echo.echo_report(
                f"Using {percent_load * 100:.0f}% of the available daemon worker slots."
            )


@workgraph_graph.command("show")
@arguments.PROCESSES()
@decorators.with_dbenv()
def process_show(processes):
    """Show details for one or multiple processes."""
    from aiida_workgraph import WorkGraph

    for process in processes:
        wg = WorkGraph.load(process.pk)
        wg.show()


@workgraph_graph.command("report")
@arguments.PROCESSES()
@click.option(
    "-i",
    "--indent-size",
    type=int,
    default=2,
    help="Set the number of spaces to indent each level by.",
)
@click.option(
    "-l",
    "--levelname",
    type=click.Choice(list(LOG_LEVELS)),
    default="REPORT",
    help="Filter the results by name of the log level.",
)
@click.option(
    "-m",
    "--max-depth",
    "max_depth",
    type=int,
    default=None,
    help="Limit the number of levels to be printed.",
)
@decorators.with_dbenv()
def process_report(processes, levelname, indent_size, max_depth):
    """Show the log report for one or multiple processes."""
    from aiida.cmdline.utils.common import (
        get_calcjob_report,
        get_process_function_report,
        get_workchain_report,
    )
    from aiida.orm import CalcFunctionNode, CalcJobNode, WorkChainNode, WorkFunctionNode

    for process in processes:
        if isinstance(process, CalcJobNode):
            echo.echo(get_calcjob_report(process))
        elif isinstance(process, WorkChainNode):
            echo.echo(get_workchain_report(process, levelname, indent_size, max_depth))
        elif isinstance(process, (CalcFunctionNode, WorkFunctionNode)):
            echo.echo(get_process_function_report(process))
        else:
            echo.echo(f"Nothing to show for node type {process.__class__}")


@workgraph_graph.command("status")
@click.option(
    "-c",
    "--call-link-label",
    "call_link_label",
    is_flag=True,
    help="Include the call link label if set.",
)
@click.option(
    "-m",
    "--max-depth",
    "max_depth",
    type=int,
    default=None,
    help="Limit the number of levels to be printed.",
)
@arguments.PROCESSES()
def process_status(call_link_label, max_depth, processes):
    """Print the status of one or multiple processes."""
    from aiida.cmdline.utils.ascii_vis import format_call_graph

    for process in processes:
        graph = format_call_graph(
            process, max_depth=max_depth, call_link_label=call_link_label
        )
        echo.echo(graph)


@workgraph_graph.command("kill")
@arguments.PROCESSES()
@options.ALL(help="Kill all processes if no specific processes are specified.")
@options.TIMEOUT()
@options.WAIT()
@decorators.with_dbenv()
def process_kill(processes, all_entries, timeout, wait):
    """Kill running processes."""
    from aiida.engine.processes import control

    if processes and all_entries:
        raise click.BadOptionUsage(
            "all",
            "cannot specify individual processes and the `--all` flag at the same time.",
        )

    if all_entries:
        click.confirm("Are you sure you want to kill all processes?", abort=True)

    try:
        message = "Killed through `verdi process kill`"
        control.kill_processes(
            processes,
            all_entries=all_entries,
            timeout=timeout,
            wait=wait,
            message=message,
        )
    except control.ProcessTimeoutException as exception:
        echo.echo_critical(f"{exception}\n{REPAIR_INSTRUCTIONS}")


@workgraph_graph.command("pause")
@arguments.PROCESSES()
@options.ALL(help="Pause all active processes if no specific processes are specified.")
@options.TIMEOUT()
@options.WAIT()
@decorators.with_dbenv()
def process_pause(processes, all_entries, timeout, wait):
    """Pause running processes."""
    from aiida.engine.processes import control

    if processes and all_entries:
        raise click.BadOptionUsage(
            "all",
            "cannot specify individual processes and the `--all` flag at the same time.",
        )

    try:
        message = "Paused through `verdi process pause`"
        control.pause_processes(
            processes,
            all_entries=all_entries,
            timeout=timeout,
            wait=wait,
            message=message,
        )
    except control.ProcessTimeoutException as exception:
        echo.echo_critical(f"{exception}\n{REPAIR_INSTRUCTIONS}")


@workgraph_graph.command("play")
@arguments.PROCESSES()
@options.ALL(help="Play all paused processes if no specific processes are specified.")
@options.TIMEOUT()
@options.WAIT()
@decorators.with_dbenv()
def process_play(processes, all_entries, timeout, wait):
    """Play (unpause) paused processes."""
    from aiida.engine.processes import control

    if processes and all_entries:
        raise click.BadOptionUsage(
            "all",
            "cannot specify individual processes and the `--all` flag at the same time.",
        )

    try:
        control.play_processes(
            processes, all_entries=all_entries, timeout=timeout, wait=wait
        )
    except control.ProcessTimeoutException as exception:
        echo.echo_critical(f"{exception}\n{REPAIR_INSTRUCTIONS}")


@workgraph_graph.command("watch")
@arguments.PROCESSES()
@decorators.with_dbenv()
@decorators.only_if_daemon_running(
    echo.echo_warning, "daemon is not running, so process may not be reachable"
)
def process_watch(processes):
    """Watch the state transitions for a process."""
    from time import sleep

    from kiwipy import BroadcastFilter

    def _print(
        communicator, body, sender, subject, correlation_id
    ):  # pylint: disable=unused-argument
        """Format the incoming broadcast data into a message and echo it to stdout."""
        if body is None:
            body = "No message specified"

        if correlation_id is None:
            correlation_id = "--"

        echo.echo(f"Process<{sender}> [{subject}|{correlation_id}]: {body}")

    communicator = get_manager().get_communicator()
    echo.echo_report("watching for broadcasted messages, press CTRL+C to stop...")

    for process in processes:

        if process.is_terminated:
            echo.echo_error(f"Process<{process.pk}> is already terminated")
            continue

        communicator.add_broadcast_subscriber(
            BroadcastFilter(_print, sender=process.pk)
        )

    try:
        # Block this thread indefinitely until interrupt
        while True:
            sleep(2)
    except (SystemExit, KeyboardInterrupt):
        echo.echo("")  # add a new line after the interrupt character
        echo.echo_report("received interrupt, exiting...")
        try:
            communicator.close()
        except RuntimeError:
            pass

        # Reraise to trigger clicks builtin abort sequence
        raise



"""`verdi process` command."""
import click

from aiida_workgraph.cli.cmd_workgraph import workgraph
from aiida.cmdline.params import arguments, options
from aiida.cmdline.utils import decorators, echo
from aiida_workgraph.cli.query_workgraph import WorkGraphQueryBuilder

REPAIR_INSTRUCTIONS = """\
If one ore more processes are unreachable, you can run the following commands to try and repair them:

    verdi daemon stop
    verdi process repair
    verdi daemon start
"""


def default_projections():
    """Return list of default projections for the ``--project`` option of ``verdi process list``.

    This indirection is necessary to prevent loading the imported module which slows down tab-completion.
    """
    return WorkGraphQueryBuilder.default_projections


@workgraph.group("task")
def workgraph_task():
    """Inspect and manage processes."""


@workgraph_task.command("pause")
@arguments.PROCESS()
@click.argument("tasks", nargs=-1)
@options.TIMEOUT()
@options.WAIT()
@decorators.with_dbenv()
def task_pause(process, tasks, timeout, wait):
    """Pause task."""
    from aiida.engine.processes import control
    from aiida_workgraph.utils.control import pause_tasks

    try:
        _, msg = pause_tasks(process.pk, tasks, timeout, wait)
    except control.ProcessTimeoutException as exception:
        echo.echo_critical(f"{exception}\n{REPAIR_INSTRUCTIONS}")


@workgraph_task.command("play")
@arguments.PROCESS()
@click.argument("tasks", nargs=-1)
@options.TIMEOUT()
@options.WAIT()
@decorators.with_dbenv()
def task_play(process, tasks, timeout, wait):
    """Play task."""
    from aiida.engine.processes import control
    from aiida_workgraph.utils.control import play_tasks

    try:
        _, msg = play_tasks(process.pk, tasks, timeout, wait)
    except control.ProcessTimeoutException as exception:
        echo.echo_critical(f"{exception}\n{REPAIR_INSTRUCTIONS}")


@workgraph_task.command("skip")
@arguments.PROCESS()
@click.argument("tasks", nargs=-1)
@options.TIMEOUT()
@options.WAIT()
@decorators.with_dbenv()
def task_skip(process, tasks, timeout, wait):
    """Skip task."""
    from aiida.engine.processes import control
    from aiida_workgraph.utils.control import skip_tasks

    for task in tasks:
        try:
            skip_tasks(process.pk, task, timeout, wait)

        except control.ProcessTimeoutException as exception:
            echo.echo_critical(f"{exception}\n{REPAIR_INSTRUCTIONS}")


@workgraph_task.command("kill")
@arguments.PROCESS()
@click.argument("tasks", nargs=-1)
@options.TIMEOUT()
@options.WAIT()
@decorators.with_dbenv()
def task_kill(process, tasks, timeout, wait):
    """Kill task."""
    from aiida.engine.processes import control
    from aiida_workgraph.utils.control import kill_tasks

    print("tasks", tasks)
    try:
        kill_tasks(process.pk, tasks, timeout, wait)
    except control.ProcessTimeoutException as exception:
        echo.echo_critical(f"{exception}\n{REPAIR_INSTRUCTIONS}")



from aiida_workgraph.cli.cmd_workgraph import workgraph
import subprocess
import click
import os
import signal
from pathlib import Path


REACT_PORT = "3000"


def get_package_root():
    """Returns the root directory of the package."""
    current_file = Path(__file__)
    # Root directory of your package
    return current_file.parent


def get_pid_file_path():
    """Get the path to the PID file in the desired directory."""
    home_dir = Path.home()
    aiida_daemon_dir = home_dir / ".aiida" / "daemon"

    # Create the directory if it does not exist
    aiida_daemon_dir.mkdir(parents=True, exist_ok=True)

    return aiida_daemon_dir / "web_processes.pid"


@workgraph.group("web")
def web():
    """Commands to manage the web application (both backend and frontend)."""


@web.command()
def start():
    """Start the web application."""
    click.echo("Starting the web application...")
    pid_file_path = get_pid_file_path()

    with open(pid_file_path, "w") as pid_file:
        # Starting FastAPI backend
        backend_process = subprocess.Popen(
            [
                "uvicorn",
                "aiida_workgraph.web.backend.app.api:app",
                "--reload",
                "--port",
                "8000",
            ]
        )
        pid_file.write(f"backend:{backend_process.pid}\n")


@web.command()
def stop():
    """Stop the web application."""
    pid_file_path = get_pid_file_path()

    if not pid_file_path.exists():
        click.echo("No running web application found.")
        return

    with open(pid_file_path, "r") as pid_file:
        for line in pid_file:
            proc_name, pid = line.strip().split(":")
            try:
                os.kill(int(pid), signal.SIGTERM)
                click.echo(f"Stopped {proc_name} (PID: {pid})")
            except ProcessLookupError:
                click.echo(f"{proc_name} (PID: {pid}) not found")

    os.remove(pid_file_path)



from aiida_workgraph.task import Task


class TestAdd(Task):

    identifier: str = "workgraph.test_add"
    name = "TestAAdd"
    node_type = "CALCFUNCTION"
    catalog = "Test"

    _executor = {
        "module": "aiida_workgraph.executors.test",
        "name": "add",
    }
    args = ["x", "y"]
    kwargs = ["t"]

    def create_properties(self) -> None:
        self.properties.new("workgraph.aiida_float", "t", default=1.0)

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        inp = self.inputs.new("workgraph.aiida_float", "x")
        inp.add_property("workgraph.aiida_float", "x", default=0.0)
        inp = self.inputs.new("workgraph.aiida_float", "y")
        inp.add_property("workgraph.aiida_float", "y", default=0.0)
        self.outputs.new("workgraph.aiida_float", "sum")


class TestGreater(Task):

    identifier: str = "workgraph.test_greater"
    name = "TestGreater"
    node_type = "CALCFUNCTION"
    catalog = "Test"

    _executor = {
        "module": "aiida_workgraph.executors.test",
        "name": "greater",
    }
    kwargs = ["x", "y"]

    def create_properties(self) -> None:
        pass

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        self.inputs.new("workgraph.aiida_float", "x")
        self.inputs.new("workgraph.aiida_float", "y")
        self.outputs.new("workgraph.aiida_bool", "result")


class TestSumDiff(Task):

    identifier: str = "workgraph.test_sum_diff"
    name = "TestSumDiff"
    node_type = "CALCFUNCTION"
    catalog = "Test"

    _executor = {
        "module": "aiida_workgraph.executors.test",
        "name": "sum_diff",
    }
    args = ["x", "y"]
    kwargs = ["t"]

    def create_properties(self) -> None:
        self.properties.new("workgraph.aiida_float", "t", default=1.0)

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        inp = self.inputs.new("workgraph.aiida_float", "x")
        inp.add_property("workgraph.aiida_float", "x", default=0.0)
        inp = self.inputs.new("workgraph.aiida_float", "y")
        inp.add_property("workgraph.aiida_float", "y", default=0.0)
        self.outputs.new("workgraph.aiida_float", "sum")
        self.outputs.new("workgraph.aiida_float", "diff")


class TestArithmeticMultiplyAdd(Task):

    identifier: str = "workgraph.test_arithmetic_multiply_add"
    name = "TestArithmeticMultiplyAdd"
    node_type = "WORKCHAIN"
    catalog = "Test"

    _executor = {
        "name": "core.arithmetic.multiply_add",
        "type": "WorkflowFactory",
    }
    kwargs = ["code", "x", "y", "z"]

    def create_properties(self) -> None:
        pass

    def create_sockets(self) -> None:
        self.inputs.clear()
        self.outputs.clear()
        self.inputs.new("workgraph.any", "code")
        inp = self.inputs.new("workgraph.aiida_int", "x")
        inp.add_property("workgraph.aiida_int", "x", default=0.0)
        inp = self.inputs.new("workgraph.aiida_int", "y")
        inp.add_property("workgraph.aiida_int", "y", default=0.0)
        inp = self.inputs.new("workgraph.aiida_int", "z")
        inp.add_property("workgraph.aiida_int", "z", default=0.0)
        self.outputs.new("workgraph.aiida_int", "result")



from typing import Dict, Optional, Union, Tuple, List, Any
from aiida.orm import load_node, Node
from datetime import datetime
from dateutil import relativedelta
from dateutil.tz import tzlocal


def get_executor_source(tdata: Any) -> Tuple[bool, Optional[str]]:
    """Get the source code of the executor."""
    import inspect
    from aiida_workgraph.utils import get_executor

    executor, _ = get_executor(tdata["executor"])
    if callable(executor):
        try:
            source_lines, _ = inspect.getsourcelines(executor)
            source_code = "".join(source_lines)
            return source_code
        except (TypeError, OSError):
            source_code = tdata["executor"].get("source_code", "")
            return source_code
    else:
        return str(executor)


def get_node_recursive(links: Dict) -> Dict[str, Union[List[int], str]]:
    """Recursively get a dictionary of nodess."""
    from collections.abc import Mapping

    data = {}
    for label, value in links.items():
        if isinstance(value, Mapping):
            data.update({label: get_node_recursive(value)})
        else:
            data[label] = [value.pk, value.__class__.__name__]
    return data


def get_node_inputs(pk: Optional[int]) -> Union[str, Dict[str, Union[List[int], str]]]:
    from aiida.common.links import LinkType

    if pk is None:
        return ""

    node = load_node(pk)
    nodes_input = node.base.links.get_incoming(
        link_type=(LinkType.INPUT_CALC, LinkType.INPUT_WORK)
    )
    if nodes_input:
        result = get_node_recursive(nodes_input.nested())
    else:
        result = {}

    return result


def get_node_outputs(pk: Optional[int]) -> Union[str, Dict[str, Union[List[int], str]]]:
    from aiida.common.links import LinkType

    if pk is None:
        return ""

    node = load_node(pk)
    result = ""
    nodes_output = node.base.links.get_outgoing(
        link_type=(LinkType.CREATE, LinkType.RETURN)
    )
    if nodes_output.all():
        result = get_node_recursive(nodes_output.nested())
    else:
        result = {}

    return result


def node_to_short_json(workgraph_pk: int, tdata: Dict[str, Any]) -> Dict[str, Any]:
    """Export a node to a rete js node."""
    from aiida_workgraph.utils import get_processes_latest

    executor = get_executor_source(tdata)
    tdata_short = {
        "node_type": tdata["metadata"]["node_type"],
        "metadata": [
            ["name", tdata["name"]],
            ["node_type", tdata["metadata"]["node_type"]],
            ["identifier", tdata["identifier"]],
        ],
        "executor": executor,
    }
    process_info = get_processes_latest(workgraph_pk, tdata["name"]).get(
        tdata["name"], {}
    )
    tdata_short["process"] = process_info
    if process_info is not None:
        tdata_short["metadata"].append(["pk", process_info.get("pk")])
        tdata_short["metadata"].append(["state", process_info.get("state")])
        tdata_short["metadata"].append(["ctime", process_info.get("ctime")])
        tdata_short["metadata"].append(["mtime", process_info.get("mtime")])
        tdata_short["inputs"] = get_node_inputs(process_info.get("pk"))
        tdata_short["outputs"] = get_node_outputs(process_info.get("pk"))
    else:
        tdata_short["inputs"] = ""
        tdata_short["outputs"] = ""

    return tdata_short


def get_node_summary(node: Node) -> List[List[str]]:
    """ """
    from plumpy import ProcessState
    from aiida.orm import ProcessNode

    table = []

    if isinstance(node, ProcessNode):
        table.append(["type", node.process_label])

        try:
            process_state = ProcessState(node.process_state)
        except (AttributeError, ValueError):
            pass
        else:
            process_state_string = process_state.value.capitalize()

            if process_state == ProcessState.FINISHED and node.exit_message:
                table.append(
                    [
                        "state",
                        f"{process_state_string} [{node.exit_status}] {node.exit_message}",
                    ]
                )
            elif process_state == ProcessState.FINISHED:
                table.append(["state", f"{process_state_string} [{node.exit_status}]"])
            elif process_state == ProcessState.EXCEPTED:
                table.append(["state", f"{process_state_string} <{node.exception}>"])
            else:
                table.append(["state", process_state_string])

    else:
        table.append(["type", node.__class__.__name__])
    table.append(["pk", str(node.pk)])
    table.append(["uuid", str(node.uuid)])
    table.append(["label", node.label])
    table.append(["description", node.description])
    table.append(["ctime", node.ctime])
    table.append(["mtime", node.mtime])

    try:
        computer = node.computer
    except AttributeError:
        pass
    else:
        if computer is not None:
            table.append(["computer", f"[{node.computer.pk}] {node.computer.label}"])

    return table


def time_ago(past_time: datetime) -> str:
    # Get the current time
    now = datetime.now(tzlocal())

    # Calculate the time difference
    delta = relativedelta.relativedelta(now, past_time)

    # Format the time difference
    if delta.years > 0:
        return f"{delta.years}Y ago"
    elif delta.months > 0:
        return f"{delta.months}M ago"
    elif delta.days > 0:
        return f"{delta.days}D ago"
    elif delta.hours > 0:
        return f"{delta.hours}h ago"
    elif delta.minutes > 0:
        return f"{delta.minutes}min ago"
    else:
        return "Just now"



from typing import List, Union, Any
from aiida_workgraph.property import TaskProperty
from node_graph.serializer import SerializeJson, SerializePickle
from aiida import orm


class PropertyAny(TaskProperty, SerializePickle):
    """A new class for Any type."""

    identifier: str = "workgraph.any"

    def validate(self, _: any) -> None:
        """No validation needed."""


class PropertyInt(TaskProperty, SerializeJson):
    """A new class for integer type."""

    identifier: str = "workgraph.int"
    allowed_types = (int, str, type(None))


class PropertyFloat(TaskProperty, SerializeJson):
    """A new class for float type."""

    identifier: str = "workgraph.float"
    allowed_types = (int, float, str, type(None))


class PropertyBool(TaskProperty, SerializeJson):
    """A new class for bool type."""

    identifier: str = "workgraph.bool"
    allowed_types = (bool, int, str, type(None))


class PropertyString(TaskProperty, SerializeJson):
    """A new class for string type."""

    identifier: str = "workgraph.string"
    allowed_types = (str, type(None))

    def validate(self, value: any) -> None:
        if not isinstance(value, self.allowed_types):
            raise TypeError(
                f"Expected value of type {self.allowed_types}, got {type(value).__name__} instead."
            )


class PropertyAiiDAInt(TaskProperty, SerializeJson):
    """A new class for integer type."""

    identifier: str = "workgraph.aiida_int"
    allowed_types = (int, str, orm.Int, type(None))

    def set_value(self, value: Union[int, orm.Int, str] = None) -> None:
        if isinstance(value, int):
            value = orm.Int(value)
        super().set_value(value)


class PropertyAiiDAFloat(TaskProperty, SerializeJson):
    """A new class for float type."""

    identifier: str = "workgraph.aiida_float"
    allowed_types = (int, float, orm.Int, orm.Float, str, type(None))

    def set_value(
        self, value: Union[int, float, orm.Int, orm.Float, str] = None
    ) -> None:
        if isinstance(value, (int, float)):
            value = orm.Float(value)
        super().set_value(value)


class PropertyAiiDABool(TaskProperty, SerializeJson):
    """A new class for bool type."""

    identifier: str = "workgraph.aiida_bool"
    allowed_types = (int, bool, orm.Int, orm.Bool, str, type(None))

    def set_value(self, value: Union[int, bool, orm.Int, orm.Bool, str] = None) -> None:
        if isinstance(value, int):
            value = orm.Bool(value)
        super().set_value(value)


class PropertyAiiDAString(TaskProperty, SerializeJson):
    """A new class for string type."""

    identifier: str = "workgraph.aiida_string"
    allowed_types = (int, float, str, orm.Str, type(None))

    def set_value(self, value: Union[int, float, str, orm.Str] = None) -> None:
        if isinstance(value, (int, float)):
            value = orm.Str(value)
        super().set_value(value)

    def validate(self, value: any) -> None:
        if not isinstance(value, self.allowed_types):
            raise TypeError(
                f"Expected value of type {self.allowed_types}, got {type(value).__name__} instead."
            )


class PropertyAiiDADict(TaskProperty, SerializeJson):
    """A new class for Dict type."""

    identifier: str = "workgraph.aiida_dict"
    allowed_types = (dict, orm.Dict, str, type(None))

    def set_value(self, value: Union[dict, orm.Dict, str] = None) -> None:
        if isinstance(value, (dict)):
            value = orm.Dict(value)
        super().set_value(value)


# ====================================
class PropertyVector(TaskProperty, SerializePickle):
    """Vector property"""

    identifier: str = "workgraph.vector"
    allowed_item_types = (object, type(None))

    def __init__(self, name, description="", size=3, default=None, update=None) -> None:
        self.size = size
        default = [] if default is None else default
        super().__init__(name, description, default, update)

    def validate(self, value: Any) -> None:
        """Validate the given value based on allowed types."""
        if value is not None:
            if len(value) != self.size:
                raise ValueError(
                    f"Invalid size: Expected {self.size}, got {len(value)} instead."
                )
            for item in value:
                if not isinstance(item, self.allowed_item_types):
                    raise ValueError(
                        f"Invalid item type: Expected {self.allowed_item_types}, got {type(item)} instead."
                    )

        super().validate(value)

    def set_value(self, value: List) -> None:
        self.validate(value)
        self._value = value
        if self.update:
            self.update()

    def copy(self):
        p = self.__class__(
            self.name, self.description, self.size, self.value, self.update
        )
        p.value = self.value
        return p

    def get_metadata(self):
        metadata = {"default": self.default, "size": self.size}
        return metadata


class PropertyAiiDAIntVector(PropertyVector):
    """A new class for integer vector type."""

    identifier: str = "workgraph.aiida_int_vector"
    allowed_types = (list, orm.List, type(None))
    allowed_item_types = (int, str, type(None))


class PropertyAiiDAFloatVector(PropertyVector):
    """A new class for float vector type."""

    identifier: str = "workgraph.aiida_float_vector"
    allowed_types = (list, orm.List, type(None))
    allowed_item_types = (int, float, str, type(None))


class PropertyStructureData(TaskProperty, SerializePickle):
    """A new class for Any type."""

    identifier: str = "workgraph.aiida_structuredata"
    allowed_types = (orm.StructureData, str, type(None))


__all__ = [
    PropertyAny,
    PropertyAiiDAInt,
    PropertyAiiDAFloat,
    PropertyAiiDABool,
    PropertyAiiDAString,
    PropertyAiiDADict,
    PropertyAiiDAIntVector,
    PropertyAiiDAFloatVector,
]



try:
    import anywidget  # noqa: F401

    USE_WIDGET = True
except ImportError:
    USE_WIDGET = False

from .workgraph import WorkGraph
from .task import Task
from .decorator import task, build_task

__version__ = "0.4.4"

__all__ = ["WorkGraph", "Task", "task", "build_task"]



from node_graph.utils import get_entries

# should after task_list, otherwise circular import
task_pool = {
    **get_entries(entry_point_name="aiida_workgraph.task"),
    **get_entries(entry_point_name="node_graph.task"),
}



from fastapi import APIRouter, HTTPException, Query
from aiida import orm
from typing import List, Dict, Union
import time

router = APIRouter()


@router.get("/api/workgraph-data")
async def read_workgraph_data(search: str = Query(None)):
    from aiida_workgraph.cli.query_workgraph import WorkGraphQueryBuilder

    try:
        relationships = {}
        builder = WorkGraphQueryBuilder()
        query_set = builder.get_query_set(
            relationships=relationships,
            # Add logic to apply search filter if search query is present
            # This is just an example, you'll need to adjust it based on your actual query builder and data model
            filters={"attributes.process_label": {"like": f"%{search}%"}}
            if search
            else None,
            # order_by, past_days, limit, etc. can also be parameters
        )
        project = ["pk", "uuid", "state", "ctime", "mtime", "process_label"]
        projected = builder.get_projected(query_set, projections=project)
        # pop headers
        projected.pop(0)
        data = []
        for p in projected:
            data.append({project[i]: p[i] for i in range(len(project))})
        return data
    except KeyError:
        raise HTTPException(status_code=404, detail=f"Workgraph {id} not found")


@router.get("/api/workgraph/{id}/{node_name}")
async def read_workgraph_task(id: int, node_name: str):
    from .utils import node_to_short_json
    from aiida.orm.utils.serialize import deserialize_unsafe
    from aiida.orm import QueryBuilder
    from aiida_workgraph.engine.workgraph import WorkGraphEngine

    try:
        tstart = time.time()
        qb = QueryBuilder()
        projections = [f"extras._workgraph.tasks.{node_name}"]
        qb.append(
            WorkGraphEngine, filters={"id": id}, project=projections, tag="process"
        )
        results = qb.all()
        if len(results) == 0:
            print("No workgraph data found in the node.")
            return
        ndata = deserialize_unsafe(results[0][0])
        content = node_to_short_json(id, ndata)
        print(f"Time to convert to json: {time.time() - tstart}")
        tstart = time.time()
        return content
    except KeyError:
        raise HTTPException(
            status_code=404, detail=f"Workgraph {id}/{node_name} not found"
        )


@router.get("/api/workgraph/{id}")
async def read_workgraph(id: int):
    from .utils import (
        get_node_summary,
        get_node_inputs,
        get_node_outputs,
    )
    from aiida_workgraph.utils import get_parent_workgraphs

    try:
        tstart = time.time()

        node = orm.load_node(id)

        content = node.base.extras.get("_workgraph_short", None)
        if content is None:
            print("No workgraph data found in the node.")
            return
        summary = {
            "table": get_node_summary(node),
            "inputs": get_node_inputs(id),
            "outputs": get_node_outputs(id),
        }

        parent_workgraphs = get_parent_workgraphs(id)
        parent_workgraphs.reverse()
        print(f"Time to load process latest: {time.time() - tstart}")
        tstart = time.time()
        content["summary"] = summary
        content["parent_workgraphs"] = parent_workgraphs
        content["processes_info"] = {}
        return content
    except KeyError:
        raise HTTPException(status_code=404, detail=f"Workgraph {id} not found")


@router.get("/api/workgraph-state/{id}")
async def read_workgraph_tasks_state(id: int, item_type: str = "task"):
    from aiida_workgraph.utils import get_processes_latest

    try:
        processes_info = get_processes_latest(id, item_type=item_type)
        return processes_info
    except KeyError:
        raise HTTPException(status_code=404, detail=f"Workgraph {id} not found")


@router.get("/api/workgraph-logs/{id}")
async def read_workgraph_logs(id: int):
    from aiida.cmdline.utils.common import get_workchain_report

    try:
        node = orm.load_node(id)
        report = get_workchain_report(node, "REPORT")
        logs = report.splitlines()
        return logs
    except KeyError:
        raise HTTPException(status_code=404, detail=f"Workgraph {id} not found")


# Route for pausing a workgraph item
@router.post("/api/workgraph/pause/{id}")
async def pause_workgraph(
    id: int,
):
    from aiida.engine.processes.control import pause_processes

    try:
        # Perform the pause action here
        node = orm.load_node(id)
        pause_processes([node])
        return {"message": f"Send message to pause workgraph {id}"}
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


# Route for playing a workgraph item
@router.post("/api/workgraph/play/{id}")
async def play_workgraph(
    id: int,
):
    from aiida.engine.processes.control import play_processes

    try:
        # Perform the play action here
        node = orm.load_node(id)
        play_processes([node])
        return {"message": f"Send message to play workgraph {id}"}
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


# Route for deleting a workgraph item
@router.delete("/api/workgraph/delete/{id}")
async def delete_workgraph(
    id: int,
    dry_run: bool = False,
) -> Dict[str, Union[bool, str, List[int]]]:
    from aiida.tools import delete_nodes

    try:
        # Perform the delete action here
        deleted_nodes, was_deleted = delete_nodes([id], dry_run=dry_run)
        if was_deleted:
            return {
                "deleted": True,
                "message": f"Deleted workgraph {id}",
                "deleted_nodes": list(deleted_nodes),
            }
        else:
            message = f"Did not delete workgraph {id}"
            if dry_run:
                message += " [dry-run]"
            return {
                "deleted": False,
                "message": message,
                "deleted_nodes": list(deleted_nodes),
            }
    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


# General function to manage task actions
async def manage_task_action(action: str, id: int, tasks: List[str]):
    from aiida_workgraph.utils.control import pause_tasks, play_tasks, kill_tasks

    print(f"Performing {action} action on tasks {tasks} in workgraph {id}")
    try:

        if action == "pause":
            print(f"Pausing tasks {tasks}")
            _, msg = pause_tasks(id, tasks=tasks)
        elif action == "play":
            print(f"Playing tasks {tasks}")
            _, msg = play_tasks(id, tasks)
        elif action == "kill":
            print(f"Killing tasks {tasks}")
            _, msg = kill_tasks(id, tasks)
        else:
            raise HTTPException(status_code=400, detail="Unsupported action")

        return {"message": msg}

    except Exception as e:
        raise HTTPException(status_code=500, detail=str(e))


# Endpoint for pausing tasks in a workgraph
@router.post("/api/workgraph/tasks/pause/{id}")
async def pause_workgraph_tasks(id: int, tasks: List[str] = None):
    return await manage_task_action("pause", id, tasks)


# Endpoint for playing tasks in a workgraph
@router.post("/api/workgraph/tasks/play/{id}")
async def play_workgraph_tasks(id: int, tasks: List[str] = None):
    return await manage_task_action("play", id, tasks)


# Endpoint for killing tasks in a workgraph
@router.post("/api/workgraph/tasks/kill/{id}")
async def kill_workgraph_tasks(id: int, tasks: List[str] = None):
    return await manage_task_action("kill", id, tasks)



from .general_data import GeneralData
from aiida import orm, common
from importlib.metadata import entry_points
from typing import Any
from aiida_workgraph.config import load_config
import sys


def get_serializer_from_entry_points() -> dict:
    """Retrieve the serializer from the entry points."""
    # import time

    # ts = time.time()
    configs = load_config()
    serializers = configs.get("serializers", {})
    excludes = serializers.get("excludes", [])
    # Retrieve the entry points for 'aiida.data' and store them in a dictionary
    eps = entry_points()
    if sys.version_info >= (3, 10):
        group = eps.select(group="aiida.data")
    else:
        group = eps.get("aiida.data", [])
    eps = {}
    for ep in group:
        # split the entry point name by first ".", and check the last part
        key = ep.name.split(".", 1)[-1]
        # skip key without "." because it is not a module name for a data type
        if "." not in key or key in excludes:
            continue
        eps.setdefault(key, [])
        eps[key].append(ep)

    # print("Time to load entry points: ", time.time() - ts)
    # check if there are duplicates
    for key, value in eps.items():
        if len(value) > 1:
            if key in serializers:
                [ep for ep in value if ep.name == serializers[key]]
                eps[key] = [ep for ep in value if ep.name == serializers[key]]
                if not eps[key]:
                    raise ValueError(
                        f"Entry point {serializers[key]} not found for {key}"
                    )
            else:
                msg = f"Duplicate entry points for {key}: {[ep.name for ep in value]}"
                raise ValueError(msg)
    return eps


eps = get_serializer_from_entry_points()


def serialize_to_aiida_nodes(inputs: dict = None) -> dict:
    """Serialize the inputs to a dictionary of AiiDA data nodes.

    Args:
        inputs (dict): The inputs to be serialized.

    Returns:
        dict: The serialized inputs.
    """
    new_inputs = {}
    # save all kwargs to inputs port
    for key, data in inputs.items():
        new_inputs[key] = general_serializer(data)
    return new_inputs


def clean_dict_key(data):
    """Replace "." with "__dot__" in the keys of a dictionary."""
    if isinstance(data, dict):
        return {k.replace(".", "__dot__"): clean_dict_key(v) for k, v in data.items()}
    return data


def general_serializer(data: Any, check_value=True) -> orm.Node:
    """Serialize the data to an AiiDA data node."""
    if isinstance(data, orm.Data):
        if check_value and not hasattr(data, "value"):
            raise ValueError("Only AiiDA data Node with a value attribute is allowed.")
        return data
    elif isinstance(data, common.extendeddicts.AttributeDict):
        # if the data is an AttributeDict, use it directly
        return data
    # if is string with syntax {{}}, this is a port will read data from ctx
    elif isinstance(data, str) and data.startswith("{{") and data.endswith("}}"):
        return data
    # if data is a class instance, get its __module__ and class name as a string
    # for example, an Atoms will have ase.atoms.Atoms
    else:
        data = clean_dict_key(data)
        # try to get the serializer from the entry points
        data_type = type(data)
        ep_key = f"{data_type.__module__}.{data_type.__name__}"
        # search for the key in the entry points
        if ep_key in eps:
            try:
                new_node = eps[ep_key][0].load()(data)
            except Exception as e:
                raise ValueError(f"Error in serializing {ep_key}: {e}")
            finally:
                # try to save the node to da
                try:
                    new_node.store()
                    return new_node
                except Exception:
                    # try to serialize the value as a GeneralData
                    try:
                        new_node = GeneralData(data)
                        new_node.store()
                        return new_node
                    except Exception as e:
                        raise ValueError(f"Error in serializing {ep_key}: {e}")
        else:
            # try to serialize the data as a GeneralData
            try:
                new_node = GeneralData(data)
                new_node.store()
                return new_node
            except Exception as e:
                raise ValueError(f"Error in serializing {ep_key}: {e}")



import aiida.orm
import node_graph
import aiida
import node_graph.link
from aiida_workgraph.socket import NodeSocket
from aiida_workgraph import USE_WIDGET
from aiida_workgraph.utils.message import WIDGET_INSTALLATION_MESSAGE
from aiida_workgraph.tasks import task_pool
from aiida_workgraph.task import Task
import time
from aiida_workgraph.collection import TaskCollection
from aiida_workgraph.utils.graph import (
    task_deletion_hook,
    task_creation_hook,
    link_creation_hook,
    link_deletion_hook,
)
from typing import Any, Dict, List, Optional, Union

if USE_WIDGET:
    from aiida_workgraph.widget import NodeGraphWidget


class WorkGraph(node_graph.NodeGraph):
    """Build flexible workflows with AiiDA.

    The class extends from NodeGraph and provides methods to run,
    submit tasks, wait for tasks to finish, and update the process status.
    It is used to handle various states of a workgraph process and provides
    convenient operations to interact with it.

    Attributes:
        process (aiida.orm.ProcessNode): The process node that represents the process status and other details.
        state (str): The current state of the workgraph process.
        pk (int): The primary key of the process node.
    """

    node_pool = task_pool

    def __init__(self, name: str = "WorkGraph", **kwargs) -> None:
        """
        Initialize a WorkGraph instance.

        Args:
            name (str, optional): The name of the WorkGraph. Defaults to 'WorkGraph'.
            **kwargs: Additional keyword arguments to be passed to the WorkGraph class.
        """
        super().__init__(name, **kwargs)
        self.context = {}
        self.workgraph_type = "NORMAL"
        self.sequence = []
        self.conditions = []
        self.process = None
        self.restart_process = None
        self.max_number_jobs = 1000000
        self.execution_count = 0
        self.max_iteration = 1000000
        self.nodes = TaskCollection(self, pool=self.node_pool)
        self.nodes.post_deletion_hooks = [task_deletion_hook]
        self.nodes.post_creation_hooks = [task_creation_hook]
        self.links.post_creation_hooks = [link_creation_hook]
        self.links.post_deletion_hooks = [link_deletion_hook]
        self._error_handlers = {}
        self._widget = NodeGraphWidget(parent=self) if USE_WIDGET else None

    @property
    def tasks(self) -> TaskCollection:
        """Add alias to `nodes` for WorkGraph"""
        return self.nodes

    def prepare_inputs(self, metadata: Optional[Dict[str, Any]]) -> Dict[str, Any]:
        from aiida_workgraph.utils import (
            merge_properties,
            serialize_properties,
        )

        wgdata = self.to_dict()
        merge_properties(wgdata)
        serialize_properties(wgdata)
        metadata = metadata or {}
        inputs = {"wg": wgdata, "metadata": metadata}
        return inputs

    def run(
        self,
        inputs: Optional[Dict[str, Any]] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Any:
        """
        Run the AiiDA workgraph process and update the process status. The method uses AiiDA's engine to run
        the process, when the process is finished, update the status of the tasks
        """
        from aiida_workgraph.engine.workgraph import WorkGraphEngine

        # set task inputs
        if inputs is not None:
            for name, input in inputs.items():
                if name not in self.tasks.keys():
                    raise KeyError(f"Task {name} not found in WorkGraph.")
                self.tasks[name].set(input)
        # One can not run again if the process is alreay created. otherwise, a new process node will
        # be created again.
        if self.process is not None:
            print("Your workgraph is already created. Please use the submit() method.")
            return
        inputs = self.prepare_inputs(metadata=metadata)
        result, node = aiida.engine.run_get_node(WorkGraphEngine, inputs=inputs)
        self.process = node
        self.update()
        return result

    def submit(
        self,
        inputs: Optional[Dict[str, Any]] = None,
        wait: bool = False,
        timeout: int = 60,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> aiida.orm.ProcessNode:
        """Submit the AiiDA workgraph process and optionally wait for it to finish.
        Args:
            wait (bool): Wait for the process to finish.
            timeout (int): The maximum time in seconds to wait for the process to finish. Defaults to 60.
            restart (bool): Restart the process, and reset the modified tasks, then only re-run the modified tasks.
            new (bool): Submit a new process.
        """
        # set task inputs
        if inputs is not None:
            for name, input in inputs.items():
                if name not in self.tasks.keys():
                    raise KeyError(f"Task {name} not found in WorkGraph.")
                self.tasks[name].set(input)

        # save the workgraph to the process node
        self.save(metadata=metadata)
        if self.process.process_state.value.upper() not in ["CREATED"]:
            raise ValueError(f"Process {self.process.pk} has already been submitted.")
        self.continue_process()
        # as long as we submit the process, it is a new submission, we should set restart_process to None
        self.restart_process = None
        if wait:
            self.wait(timeout=timeout)
        return self.process

    def save(self, metadata: Optional[Dict[str, Any]] = None) -> None:
        """Save the udpated workgraph to the process
        This is only used for a running workgraph.
        Save the AiiDA workgraph process and update the process status.
        """
        from aiida.manage import manager
        from aiida.engine.utils import instantiate_process
        from aiida_workgraph.engine.workgraph import WorkGraphEngine

        inputs = self.prepare_inputs(metadata)
        if self.process is None:
            runner = manager.get_manager().get_runner()
            # init a process node
            process_inited = instantiate_process(runner, WorkGraphEngine, **inputs)
            process_inited.runner.persister.save_checkpoint(process_inited)
            self.process = process_inited.node
            self.process_inited = process_inited
            process_inited.close()
            print(f"WorkGraph process created, PK: {self.process.pk}")
        else:
            self.save_to_base(inputs["wg"])
        self.update()

    def save_to_base(self, wgdata: Dict[str, Any]) -> None:
        """Save new wgdata to base.extras.
        It will first check the difference, and reset tasks if needed.
        """
        from aiida_workgraph.utils.analysis import WorkGraphSaver

        saver = WorkGraphSaver(
            self.process, wgdata, restart_process=self.restart_process
        )
        saver.save()

    def to_dict(self, store_nodes=False) -> Dict[str, Any]:
        from aiida_workgraph.utils import store_nodes_recursely

        wgdata = super().to_dict()
        # save the sequence and context
        self.context["_sequence"] = self.sequence
        # only alphanumeric and underscores are allowed
        wgdata["context"] = {
            key.replace(".", "__"): value for key, value in self.context.items()
        }
        wgdata.update(
            {
                "restart_process": aiida.orm.Int(self.restart_process.pk)
                if self.restart_process
                else None,
                "max_iteration": self.max_iteration,
                "execution_count": self.execution_count,
                "workgraph_type": self.workgraph_type,
                "conditions": self.conditions,
                "max_number_jobs": self.max_number_jobs,
            }
        )
        # save error handlers
        wgdata["error_handlers"] = self.get_error_handlers()
        wgdata["tasks"] = wgdata.pop("nodes")
        if store_nodes:
            store_nodes_recursely(wgdata)
        return wgdata

    def get_error_handlers(self) -> Dict[str, Any]:
        """Get the error handlers."""
        from aiida.engine import ExitCode

        from aiida_workgraph.utils import build_callable

        error_handlers = {}
        for name, error_handler in self._error_handlers.items():
            error_handler["handler"] = build_callable(error_handler["handler"])
            error_handlers[name] = error_handler
        # convert exit code label (str) to status (int)
        for handler in error_handlers.values():
            for task in handler["tasks"].values():
                exit_codes = []
                for exit_code in task["exit_codes"]:
                    if isinstance(exit_code, int):
                        exit_codes.append(exit_code)
                    elif isinstance(exit_code, ExitCode):
                        exit_codes.append(exit_code.status)
                    else:
                        raise ValueError(
                            f"Exit code {exit_code} is not a valid exit code."
                        )
                task["exit_codes"] = exit_codes
        return error_handlers

    def wait(self, timeout: int = 50, tasks: dict = None) -> None:
        """
        Periodically checks and waits for the AiiDA workgraph process to finish until a given timeout.
        Args:
            timeout (int): The maximum time in seconds to wait for the process to finish. Defaults to 50.
        """
        terminating_states = (
            "KILLED",
            "PAUSED",
            "FINISHED",
            "FAILED",
            "CANCELLED",
            "EXCEPTED",
        )
        start = time.time()
        self.update()
        finished = False
        while not finished:
            self.update()
            if tasks is not None:
                states = []
                for name, value in tasks.items():
                    flag = self.tasks[name].state in value
                    states.append(flag)
                finished = all(states)
            else:
                finished = self.state in terminating_states
            time.sleep(0.5)
            if time.time() - start > timeout:
                break

    def update(self) -> None:
        """
        Update the current state and primary key of the process node as well as the state, node and primary key
        of the tasks that are outgoing from the process node. This includes updating the state of process nodes
        linked to the current process, and data nodes linked to the current process.
        """
        # from aiida_workgraph.utils import get_executor
        from aiida_workgraph.utils import get_nested_dict, get_processes_latest

        if self.process is None:
            return

        self.state = self.process.process_state.value.upper()
        processes_data = get_processes_latest(self.pk)
        for name, data in processes_data.items():
            self.tasks[name].state = data["state"]
            self.tasks[name].ctime = data["ctime"]
            self.tasks[name].mtime = data["mtime"]
            self.tasks[name].pk = data["pk"]
            if data["pk"] is not None:
                node = aiida.orm.load_node(data["pk"])
                self.tasks[name].process = self.tasks[name].node = node
                if isinstance(node, aiida.orm.ProcessNode) and getattr(
                    node, "process_state", False
                ):
                    # if the node is finished ok, update the output sockets
                    # note the task.state may not be the same as the node.process_state
                    # for example, task.state can be `SKIPPED` if it is inside a conditional block,
                    # even if the node.is_finished_ok is True
                    if node.is_finished_ok:
                        # update the output sockets
                        i = 0
                        for socket in self.tasks[name].outputs:
                            socket.value = get_nested_dict(
                                node.outputs, socket.name, default=None
                            )
                            i += 1
                # read results from the process outputs
                elif isinstance(node, aiida.orm.Data):
                    self.tasks[name].outputs[0].value = node
        execution_count = getattr(self.process.outputs, "execution_count", None)
        self.execution_count = execution_count if execution_count else 0
        if self._widget is not None:
            states = {name: data["state"] for name, data in processes_data.items()}
            self._widget.states = states

    @property
    def pk(self) -> Optional[int]:
        return self.process.pk if self.process else None

    @classmethod
    def from_dict(cls, wgdata: Dict[str, Any]) -> "WorkGraph":

        if "tasks" in wgdata:
            wgdata["nodes"] = wgdata.pop("tasks")
        wg = super().from_dict(wgdata)
        for key in [
            "max_iteration",
            "execution_count",
            "workgraph_type",
            "conditions",
            "max_number_jobs",
        ]:
            if key in wgdata:
                setattr(wg, key, wgdata[key])
        if "error_handlers" in wgdata:
            wg._error_handlers = wgdata["error_handlers"]
        return wg

    @classmethod
    def from_yaml(cls, filename: str = None, string: str = None) -> "WorkGraph":
        """Build WrokGraph from yaml file."""
        import yaml
        from node_graph.utils import yaml_to_dict

        # load data
        if filename:
            with open(filename, "r") as f:
                wgdata = yaml.safe_load(f)
        elif string:
            wgdata = yaml.safe_load(string)
        else:
            raise Exception("Please specific a filename or yaml string.")
        wgdata["nodes"] = wgdata.pop("tasks")
        wgdata = yaml_to_dict(wgdata)
        nt = cls.from_dict(wgdata)
        return nt

    @classmethod
    def load(cls, pk: int) -> Optional["WorkGraph"]:
        """
        Load WorkGraph from the process node with the given primary key.

        Args:
            pk (int): The primary key of the process node.
        """
        from aiida_workgraph.utils import get_workgraph_data

        process = aiida.orm.load_node(pk)
        wgdata = get_workgraph_data(process)
        if wgdata is None:
            print("No workgraph data found in the process node.")
            return
        wg = cls.from_dict(wgdata)
        wg.process = process
        wg.update()
        return wg

    def show(self) -> None:
        """
        Print the current state of the workgraph process.
        """
        from tabulate import tabulate

        table = []
        self.update()
        for task in self.tasks:
            table.append([task.name, task.pk, task.state])
        print("-" * 80)
        print("WorkGraph: {}, PK: {}, State: {}".format(self.name, self.pk, self.state))
        print("-" * 80)
        print("Tasks:")
        print(tabulate(table, headers=["Name", "PK", "State"]))
        print("-" * 80)

    def pause(self) -> None:
        """Pause the workgraph."""
        # from aiida.engine.processes import control
        # try:
        # control.pause_processes([self.process])
        import os

        os.system("verdi process pause {}".format(self.process.pk))

    def pause_tasks(self, tasks: List[str]) -> None:
        """Pause the given tasks."""
        from aiida_workgraph.utils.control import pause_tasks

        self.update()

        if self.process is None:
            for name in tasks:
                self.tasks[name].action = "PAUSE"
        else:
            _, msg = pause_tasks(self.process.pk, tasks)

        return "Send message to pause tasks."

    def play_tasks(self, tasks: List[str]) -> None:
        """Play the given tasks"""

        from aiida_workgraph.utils.control import play_tasks

        if self.process is None:
            for name in tasks:
                self.tasks[name].action = ""
        else:
            _, msg = play_tasks(self.process.pk, tasks)
        return "Send message to play tasks."

    def kill_tasks(self, tasks: List[str]) -> None:
        """Kill the given tasks"""

        from aiida_workgraph.utils.control import kill_tasks

        if self.process is None:
            for name in tasks:
                self.tasks[name].action = "KILL"
        else:
            _, msg = kill_tasks(self.process.pk, tasks)
        return "Send message to kill tasks."

    def continue_process(self):
        """Continue a saved process by sending the task to RabbitMA.
        Use with caution, this may launch duplicate processes."""
        from aiida.manage import get_manager

        process_controller = get_manager().get_process_controller()
        process_controller.continue_process(self.pk)

    def play(self):
        import os

        os.system("verdi process play {}".format(self.process.pk))

    def restart(self):
        """Create a restart submission."""
        if self.process is None:
            raise ValueError(
                "No process found. One can not restart from a non-existing process."
            )
        # save the current process node as restart_process
        # so that the WorkGraphSaver can compare the difference, and reset the modified tasks
        self.restart_process = self.process
        self.process = None

    def reset(self) -> None:
        """Reset the workgraph to create a new submission."""

        self.process = None
        for task in self.tasks:
            task.reset()
        self.sequence = []
        self.conditions = []
        self.context = {}
        self.state = "PLANNED"

    def extend(self, wg: "WorkGraph", prefix: str = "") -> None:
        """Append a workgraph to the current workgraph.
        prefix is used to add a prefix to the task names.
        """
        for task in wg.tasks:
            task.name = prefix + task.name
            task.parent = self
            self.tasks.append(task)
        # self.sequence.extend([prefix + task for task in wg.sequence])
        # self.conditions.extend(wg.conditions)
        self.context.update(wg.context)
        # links
        for link in wg.links:
            self.links.append(link)

    @property
    def error_handlers(self) -> Dict[str, Any]:
        """Get the error handlers."""
        return self._error_handlers

    def add_error_handler(self, handler, name, tasks: dict = None) -> None:
        """Attach an error handler to the workgraph."""
        self._error_handlers[name] = {"handler": handler, "tasks": tasks}

    def _repr_mimebundle_(self, *args, **kwargs):

        if self._widget is None:
            print(WIDGET_INSTALLATION_MESSAGE)
            return

        # if ipywdigets > 8.0.0, use _repr_mimebundle_ instead of _ipython_display_
        self._widget.from_workgraph(self)
        if hasattr(self._widget, "_repr_mimebundle_"):
            return self._widget._repr_mimebundle_(*args, **kwargs)
        else:
            return self._widget._ipython_display_(*args, **kwargs)

    def add_task(
        self, identifier: Union[str, callable], name: str = None, **kwargs
    ) -> Task:
        """Add a task to the workgraph."""
        node = self.tasks.new(identifier, name, **kwargs)
        return node

    def add_link(
        self, source: NodeSocket, target: NodeSocket
    ) -> node_graph.link.NodeLink:
        """Add a link between two nodes."""
        link = self.links.new(source, target)
        return link

    def to_html(self, output: str = None, **kwargs):
        """Write a standalone html file to visualize the workgraph."""
        if self._widget is None:
            print(WIDGET_INSTALLATION_MESSAGE)
            return
        self._widget.from_workgraph(self)
        return self._widget.to_html(output=output, **kwargs)



async def monitor(function, interval, timeout, *args, **kwargs):
    """Monitor the function until it returns True or the timeout is reached."""
    import asyncio
    import time

    start_time = time.time()
    while True:
        result = function(*args, **kwargs)
        if result:
            break
        if time.time() - start_time > timeout:
            raise TimeoutError(f"Timeout reached for monitor function {function}")
        await asyncio.sleep(interval)


def file_monitor(filename: str):
    """Check if the file exists."""
    import os

    return os.path.exists(filename)


def time_monitor(time: str):
    """Return True if the current time is greater than the given time."""
    import datetime

    # load the time string
    time = datetime.datetime.strptime(time, "%Y-%m-%d %H:%M:%S.%f")

    return datetime.datetime.now() > time


def task_monitor(task_name: str, workgraph_pk: int = None, workgraph_name: str = None):
    """Return True if the task in the WorkGraph is completed."""
    from aiida import orm
    from aiida_workgraph.engine.workgraph import WorkGraphEngine

    if workgraph_pk:
        try:
            node = orm.load_node(workgraph_pk)
        except Exception:
            return False
    else:
        builder = orm.QueryBuilder()
        builder.append(
            WorkGraphEngine,
            filters={
                "attributes.process_label": {"==": f"WorkGraph<{workgraph_name}>"}
            },
            tag="process",
        )
        if builder.count() == 0:
            return False
    print("Found workgraph")
    node = builder.first()[0]
    state = node.base.extras.get(f"_task_state_{task_name}", None)
    print(f"Task state: {state}")
    if state in ["FINISHED", "FAILED", "SKIPPED"]:
        return True
    else:
        return False



from typing import Optional, Dict, Tuple, List

# import datetime
from aiida.orm import ProcessNode
from aiida.orm.utils.serialize import serialize, deserialize_unsafe


class WorkGraphSaver:
    """Save a workgraph to the database."""

    def __init__(
        self,
        process: ProcessNode,
        wgdata: Dict,
        restart_process: Optional[ProcessNode] = None,
    ) -> None:
        """Init WorkGraphSaver.

        Args:
            wgdata (dict): data of workgraph to be launched.
        """
        self.process = process
        self.restart_process = restart_process
        self.wgdata = wgdata
        self.uuid = wgdata["uuid"]
        self.name = wgdata["name"]
        self.wait_to_link()
        self.clean_hanging_links()

    def wait_to_link(self) -> None:
        """Convert wait attribute to link."""
        for name, task in self.wgdata["tasks"].items():
            for wait_task in task["wait"]:
                if wait_task in self.wgdata["tasks"]:
                    self.wgdata["links"].append(
                        {
                            "from_node": wait_task,
                            "from_socket": "_wait",
                            "to_node": name,
                            "to_socket": "_wait",
                        }
                    )
                else:
                    raise ValueError(
                        "Task {} wait for a non-exist task {}".format(name, wait_task)
                    )

    def clean_hanging_links(self) -> None:
        """Clean hanging links in the workgraph."""
        for link in self.wgdata["links"][:]:  # Iterate over a shallow copy of the list
            if (
                link["from_node"] not in self.wgdata["tasks"]
                or link["to_node"] not in self.wgdata["tasks"]
            ):
                self.wgdata["links"].remove(link)

    def save(self) -> None:
        """Save workgraph.

        - Update uuid for links. Build compressed tasks for workgraph.
        - Analysis connectivity
        - Check exist in database or not. If not in database, save directly.
        - If in database, analyze the difference, save accordingly.
        """
        self.build_task_link()
        self.assign_zone()
        self.build_connectivity()
        self.update_parent_task()
        self.find_all_zones_inputs()
        if self.exist_in_db() or self.restart_process is not None:
            new_tasks, modified_tasks, update_metadata = self.check_diff(
                self.restart_process
            )
            print("modified_tasks: {}".format(modified_tasks))
            self.reset_tasks(modified_tasks)
        self.insert_workgraph_to_db()

    def build_task_link(self) -> None:
        """Create links for tasks.
        Create the links for task inputs using:
        1) workgraph links

        """
        # reset task input links
        for name, task in self.wgdata["tasks"].items():
            for _, input in task["inputs"].items():
                input["links"] = []
            for _, output in task["outputs"].items():
                output["links"] = []
        for link in self.wgdata["links"]:
            to_socket = [
                socket
                for name, socket in self.wgdata["tasks"][link["to_node"]][
                    "inputs"
                ].items()
                if name == link["to_socket"]
            ][0]
            from_socket = [
                socket
                for name, socket in self.wgdata["tasks"][link["from_node"]][
                    "outputs"
                ].items()
                if name == link["from_socket"]
            ][0]
            to_socket["links"].append(link)
            from_socket["links"].append(link)

    def assign_zone(self) -> None:
        """Assign zone for each task."""
        # assign parent_task for each task
        for name, task in self.wgdata["tasks"].items():
            for child_task in task["children"]:
                self.wgdata["tasks"][child_task]["parent_task"][0] = name

    def update_parent_task(self) -> None:
        """Recursively update the list of parent tasks for each task in wgdata."""

        def get_all_parents(task_name):
            """Recursively collect all parent tasks for a given task."""
            task = self.wgdata["tasks"][task_name]
            parent_names = []
            while task["parent_task"][0] is not None:
                parent_task_name = task["parent_task"][0]
                parent_names.append(parent_task_name)
                task = self.wgdata["tasks"][parent_task_name]
            parent_names.append(None)
            return parent_names

        # Loop through all tasks and update their parent task lists recursively
        for name, task in self.wgdata["tasks"].items():
            task["parent_task"] = get_all_parents(name)

    def find_all_zones_inputs(self) -> None:
        for name in self.wgdata["tasks"]:
            self.find_zone_inputs(name)

    def find_zone_inputs(self, name: str) -> None:
        """Find the input and outputs tasks for the zone."""
        task = self.wgdata["tasks"][name]
        input_tasks = []
        for _, input in self.wgdata["tasks"][name]["inputs"].items():
            for link in input["links"]:
                input_tasks.append(link["from_node"])
        # find all the input tasks
        for child_task in task["children"]:
            # if the child task is a zone
            if self.wgdata["tasks"][child_task]["children"]:
                # find the input tasks of the child task zone
                self.find_zone_inputs(child_task)
                # find all the input tasks which outside the while zone
                for child_task1 in self.wgdata["connectivity"]["zone"][child_task][
                    "input_tasks"
                ]:
                    if child_task1 not in task["children"]:
                        input_tasks.append(child_task1)
            else:
                # if the child task is not a zone, get the input tasks of the child task
                # find all the input tasks which outside the while zone
                for _, input in self.wgdata["tasks"][child_task]["inputs"].items():
                    for link in input["links"]:
                        input_tasks.append(link["from_node"])
        # find the input tasks which are not in the zone
        new_input_tasks = []
        for input_task in input_tasks:
            task_to_check = [input_task]
            task_to_check.extend(self.wgdata["tasks"][input_task]["parent_task"])
            not_in_the_zone = True
            for task_name in task_to_check:
                if task_name in task["children"]:
                    not_in_the_zone = False
                    break
            if not_in_the_zone:
                new_input_tasks.append(input_task)
        # find the parent task of the input tasks
        final_input_tasks = []
        parent_tasks = task["parent_task"]
        for input_task in new_input_tasks:
            # find the first parent task of this two task which are the same
            for parent_task in parent_tasks:
                if parent_task in self.wgdata["tasks"][input_task]["parent_task"]:
                    break
            # add the input task to the parent task
            if parent_task in self.wgdata["tasks"][input_task]["parent_task"]:
                index = self.wgdata["tasks"][input_task]["parent_task"].index(
                    parent_task
                )
            else:
                index = 0
            if index == 0:
                final_input_tasks.append(input_task)
            else:
                final_input_tasks.append(
                    self.wgdata["tasks"][input_task]["parent_task"][index - 1]
                )
        # remove the duplicate tasks
        final_input_tasks = list(set(final_input_tasks))
        self.wgdata["connectivity"]["zone"][name] = {"input_tasks": final_input_tasks}

    def insert_workgraph_to_db(self) -> None:
        """Save a new workgraph in the database.

        - workgraph
        - all tasks
        """
        from aiida_workgraph.utils import workgraph_to_short_json
        import inspect
        from aiida_workgraph.orm.function_data import PickledLocalFunction

        # pprint(self.wgdata)
        # self.wgdata["created"] = datetime.datetime.utcnow()
        # self.wgdata["lastUpdate"] = datetime.datetime.utcnow()
        short_wgdata = workgraph_to_short_json(self.wgdata)
        self.process.base.extras.set("_workgraph_short", short_wgdata)
        self.save_task_states()
        for name, task in self.wgdata["tasks"].items():
            for _, input in task["inputs"].items():
                if input["property"] is None:
                    continue
                prop = input["property"]
                if inspect.isfunction(prop["value"]):
                    prop["value"] = PickledLocalFunction(prop["value"]).store()
            self.wgdata["tasks"][name] = serialize(task)
        # nodes is a copy of tasks, so we need to pop it out
        self.wgdata["error_handlers"] = serialize(self.wgdata["error_handlers"])
        self.wgdata["context"] = serialize(self.wgdata["context"])
        self.process.base.extras.set("_workgraph", self.wgdata)

    def save_task_states(self) -> Dict:
        """Get task states."""

        task_states = {}
        task_processes = {}
        task_actions = {}
        for name, task in self.wgdata["tasks"].items():
            task_states[f"_task_state_{name}"] = task["state"]
            task_processes[f"_task_process_{name}"] = task["process"]
            task_actions[f"_task_action_{name}"] = task["action"]
        self.process.base.extras.set_many(task_states)
        self.process.base.extras.set_many(task_processes)
        self.process.base.extras.set_many(task_actions)

        return task_states

    def reset_tasks(self, tasks: List[str]) -> None:
        """Reset tasks

        Args:
            tasks (list): a list of task names.
        """
        from aiida_workgraph.utils.control import create_task_action

        # print("process state: ", self.process.process_state.value.upper())
        if (
            self.process.process_state is None
            or self.process.process_state.value.upper() == "CREATED"
        ):
            for name in tasks:
                self.wgdata["tasks"][name]["state"] = "PLANNED"
                self.wgdata["tasks"][name]["process"] = serialize(None)
                self.wgdata["tasks"][name]["result"] = None
                names = self.wgdata["connectivity"]["child_node"][name]
                for name in names:
                    self.wgdata["tasks"][name]["state"] = "PLANNED"
                    self.wgdata["tasks"][name]["result"] = None
                    self.wgdata["tasks"][name]["process"] = serialize(None)
        else:
            create_task_action(self.process.pk, tasks=tasks, action="reset")

    def set_tasks_action(self, action: str) -> None:
        """Set task action."""
        for name, task in self.wgdata["tasks"].items():
            # print("Reset task: {}".format(task))
            task["action"] = action

    def get_wgdata_from_db(
        self, process: Optional[ProcessNode] = None
    ) -> Optional[Dict]:

        process = self.process if process is None else process
        wgdata = process.base.extras.get("_workgraph", None)
        if wgdata is None:
            print("No workgraph data found in the process node.")
            return
        for name, task in wgdata["tasks"].items():
            wgdata["tasks"][name] = deserialize_unsafe(task)
        wgdata["error_handlers"] = deserialize_unsafe(wgdata["error_handlers"])
        return wgdata

    def check_diff(
        self, restart_process: Optional[ProcessNode] = None
    ) -> Tuple[List[str], List[str], Dict]:
        """Find difference between workgraph and its database.

        Returns:
            new_tasks: new tasks
            modified_tasks: modified tasks
        """
        from node_graph.analysis import DifferenceAnalysis

        wg1 = self.get_wgdata_from_db(restart_process)
        # change tasks to nodes for DifferenceAnalysis
        wg1["nodes"] = wg1.pop("tasks")
        self.wgdata["nodes"] = self.wgdata.pop("tasks")
        dc = DifferenceAnalysis(nt1=wg1, nt2=self.wgdata)
        (
            new_tasks,
            modified_tasks,
            update_metadata,
        ) = dc.build_difference()
        # change nodes back to tasks
        wg1["tasks"] = wg1.pop("nodes")
        self.wgdata["tasks"] = self.wgdata.pop("nodes")
        return new_tasks, modified_tasks, update_metadata

    def exist_in_db(self) -> bool:
        """Check workgraph exist in database or not.

        Returns:
            bool: _description_
        """
        if self.process.base.extras.get("_workgraph", None) is not None:
            return True
        return False

    def build_connectivity(self) -> None:
        """Analyze the connectivity of workgraph and save it into dict."""
        from node_graph.analysis import ConnectivityAnalysis

        # ConnectivityAnalysis use nodes instead of tasks
        self.wgdata["nodes"] = self.wgdata.pop("tasks")
        nc = ConnectivityAnalysis(self.wgdata)
        self.wgdata["connectivity"] = nc.build_connectivity()
        self.wgdata["connectivity"]["zone"] = {}
        # change nodes back to tasks
        self.wgdata["tasks"] = self.wgdata.pop("nodes")



import inspect
import textwrap
from typing import Callable, Dict, Any, get_type_hints, _SpecialForm
from .general_data import GeneralData


class PickledFunction(GeneralData):
    """Data class to represent a pickled Python function."""

    def __init__(self, value=None, **kwargs):
        """Initialize a PickledFunction node instance.

        :param value: a Python function
        """
        super().__init__(**kwargs)
        if not callable(value):
            raise ValueError("value must be a callable Python function")
        self.set_value(value)
        self.set_attribute(value)

    def __str__(self):
        return (
            f"PickledFunction<{self.base.attributes.get('function_name')}> pk={self.pk}"
        )

    @property
    def metadata(self):
        """Return a dictionary of metadata."""
        return {
            "name": self.base.attributes.get("name"),
            "import_statements": self.base.attributes.get("import_statements"),
            "source_code": self.base.attributes.get("source_code"),
            "source_code_without_decorator": self.base.attributes.get(
                "source_code_without_decorator"
            ),
            "type": "function",
            "is_pickle": True,
        }

    @classmethod
    def build_callable(cls, func):
        """Return the executor for this node."""
        import cloudpickle as pickle

        executor = {
            "executor": pickle.dumps(func),
            "type": "function",
            "is_pickle": True,
        }
        executor.update(cls.inspect_function(func))
        return executor

    def set_attribute(self, value):
        """Set the contents of this node by pickling the provided function.

        :param value: The Python function to pickle and store.
        """
        # Serialize the function and extract metadata
        serialized_data = self.inspect_function(value)

        # Store relevant metadata
        self.base.attributes.set("name", serialized_data["name"])
        self.base.attributes.set(
            "import_statements", serialized_data["import_statements"]
        )
        self.base.attributes.set("source_code", serialized_data["source_code"])
        self.base.attributes.set(
            "source_code_without_decorator",
            serialized_data["source_code_without_decorator"],
        )

    @classmethod
    def inspect_function(cls, func: Callable) -> Dict[str, Any]:
        """Serialize a function for storage or transmission."""
        try:
            # we need save the source code explicitly, because in the case of jupyter notebook,
            # the source code is not saved in the pickle file
            source_code = inspect.getsource(func)
            # Split the source into lines for processing
            source_code_lines = source_code.split("\n")
            function_source_code = "\n".join(source_code_lines)
            # Find the first line of the actual function definition
            for i, line in enumerate(source_code_lines):
                if line.strip().startswith("def "):
                    break
            function_source_code_without_decorator = "\n".join(source_code_lines[i:])
            function_source_code_without_decorator = textwrap.dedent(
                function_source_code_without_decorator
            )
            # we also need to include the necessary imports for the types used in the type hints.
            try:
                required_imports = cls.get_required_imports(func)
            except Exception as e:
                required_imports = {}
                print(
                    f"Failed to get required imports for function {func.__name__}: {e}"
                )
            # Generate import statements
            import_statements = "\n".join(
                f"from {module} import {', '.join(types)}"
                for module, types in required_imports.items()
            )
        except Exception as e:
            print(f"Failed to inspect function {func.__name__}: {e}")
            function_source_code = ""
            function_source_code_without_decorator = ""
            import_statements = ""
        return {
            "name": func.__name__,
            "source_code": function_source_code,
            "source_code_without_decorator": function_source_code_without_decorator,
            "import_statements": import_statements,
        }

    @classmethod
    def get_required_imports(cls, func: Callable) -> Dict[str, set]:
        """Retrieve type hints and the corresponding modules."""
        type_hints = get_type_hints(func)
        imports = {}

        def add_imports(type_hint):
            if isinstance(
                type_hint, _SpecialForm
            ):  # Handle special forms like Any, Union, Optional
                module_name = "typing"
                type_name = type_hint._name or str(type_hint)
            elif hasattr(
                type_hint, "__origin__"
            ):  # This checks for higher-order types like List, Dict
                module_name = type_hint.__module__
                type_name = getattr(type_hint, "_name", None) or getattr(
                    type_hint.__origin__, "__name__", None
                )
                for arg in getattr(type_hint, "__args__", []):
                    if arg is type(None):  # noqa: E721
                        continue
                    add_imports(arg)  # Recursively add imports for each argument
            elif hasattr(type_hint, "__module__"):
                module_name = type_hint.__module__
                type_name = type_hint.__name__
            else:
                return  # If no module or origin, we can't import it, e.g., for literals

            if type_name is not None:
                if module_name not in imports:
                    imports[module_name] = set()
                imports[module_name].add(type_name)

        for _, type_hint in type_hints.items():
            add_imports(type_hint)

        return imports


def to_pickled_function(value):
    """Convert a Python function to a `PickledFunction` instance."""
    return PickledFunction(value)


class PickledLocalFunction(PickledFunction):
    """PickledFunction subclass for local functions."""



from typing import Any, Type, Union, Dict, Optional, Callable
from node_graph.property import NodeProperty
import re


class TaskProperty(NodeProperty):
    """Represent a property of a Task in the AiiDA WorkGraph."""

    def validate(self, value: any) -> None:
        if isinstance(value, str) and not re.search(r"\{\{.*?\}\}", value):
            raise TypeError(
                f"""Expected value of type {self.allowed_types}, got {type(value).__name__} instead.
If you want to use variable from context, use double curly braces like this: {{{{variable_name}}}}"""
            )
        super().validate(value)

    @classmethod
    def new(
        cls,
        identifier: Union[Callable, str],
        name: Optional[str] = None,
        data: Dict[str, Any] = {},
    ) -> "TaskProperty":
        """Create a property from a identifier."""
        # use property_pool from aiida_workgraph.properties
        # to override the default property_pool from node_graph
        from aiida_workgraph.properties import property_pool

        # build the task on the fly if the identifier is a callable
        if callable(identifier):
            identifier = build_property_from_AiiDA(identifier)
        return super().new(
            identifier, name=name, data=data, property_pool=property_pool
        )


def build_property_from_AiiDA(DataClass: Type[Any]) -> Type[TaskProperty]:
    """Create a property class from AiiDA DataClass."""

    class AiiDATaskProperty(TaskProperty):
        identifier: str = DataClass.__name__

        def set_value(self, value: Any) -> None:
            # run the callback function
            if isinstance(value, DataClass):
                self._value = value
                if self.update is not None:
                    self.update()
            elif (
                isinstance(value, str)
                and value.rstrip().startswith("{{")
                and value.endswith("}}")
            ):
                self._value = value
            else:
                raise Exception("{} is not an {}.".format(value, DataClass.__name__))

        def get_serialize(self) -> Dict[str, str]:
            serialize = {"module": "aiida.orm.utils.serialize", "name": "serialize"}
            return serialize

        def get_deserialize(self) -> Dict[str, str]:
            deserialize = {
                "module": "aiida.orm.utils.serialize",
                "name": "deserialize_unsafe",
            }
            return deserialize

    return AiiDATaskProperty



from node_graph.collection import (
    NodeCollection,
    PropertyCollection,
    InputSocketCollection,
    OutputSocketCollection,
)
from typing import Any, Callable, Optional, Union


class TaskCollection(NodeCollection):
    def new(
        self,
        identifier: Union[Callable, str],
        name: Optional[str] = None,
        uuid: Optional[str] = None,
        **kwargs: Any
    ) -> Any:
        from aiida_workgraph.decorator import (
            build_task_from_callable,
            build_pythonjob_task,
            build_shelljob_task,
            build_task_from_workgraph,
        )
        from aiida_workgraph.workgraph import WorkGraph

        # build the task on the fly if the identifier is a callable
        if callable(identifier):
            identifier = build_task_from_callable(identifier)
        if isinstance(identifier, str) and identifier.upper() == "PYTHONJOB":
            identifier, _ = build_pythonjob_task(kwargs.pop("function"))
        if isinstance(identifier, str) and identifier.upper() == "SHELLJOB":
            identifier, _, links = build_shelljob_task(
                nodes=kwargs.get("nodes", {}),
                outputs=kwargs.get("outputs", None),
                parser_outputs=kwargs.pop("parser_outputs", None),
            )
            task = super().new(identifier, name, uuid, **kwargs)
            # make links between the tasks
            task.set(links)
            return task
        if isinstance(identifier, str) and identifier.upper() == "WHILE":
            task = super().new("workgraph.while", name, uuid, **kwargs)
            return task
        if isinstance(identifier, str) and identifier.upper() == "IF":
            task = super().new("workgraph.if", name, uuid, **kwargs)
            return task
        if isinstance(identifier, WorkGraph):
            identifier = build_task_from_workgraph(identifier)
        return super().new(identifier, name, uuid, **kwargs)


class WorkGraphPropertyCollection(PropertyCollection):
    def new(
        self,
        identifier: Union[Callable, str],
        name: Optional[str] = None,
        **kwargs: Any
    ) -> Any:
        from aiida_workgraph.property import build_property_from_AiiDA

        # build the property on the fly if the identifier is a callable
        if callable(identifier):
            identifier = build_property_from_AiiDA(identifier)
        # Call the original new method
        return super().new(identifier, name, **kwargs)


class WorkGraphInputSocketCollection(InputSocketCollection):
    def new(
        self,
        identifier: Union[Callable, str],
        name: Optional[str] = None,
        **kwargs: Any
    ) -> Any:
        from aiida_workgraph.socket import build_socket_from_AiiDA

        # build the socket on the fly if the identifier is a callable
        if callable(identifier):
            identifier = build_socket_from_AiiDA(identifier)
        # Call the original new method
        return super().new(identifier, name, **kwargs)


class WorkGraphOutputSocketCollection(OutputSocketCollection):
    def new(
        self,
        identifier: Union[Callable, str],
        name: Optional[str] = None,
        **kwargs: Any
    ) -> Any:
        from aiida_workgraph.socket import build_socket_from_AiiDA

        # build the socket on the fly if the identifier is a callable
        if callable(identifier):
            identifier = build_socket_from_AiiDA(identifier)
        # Call the original new method
        return super().new(identifier, name, **kwargs)



from typing import Any, Type

from aiida import orm
from node_graph.socket import NodeSocket

from aiida_workgraph.property import TaskProperty


class TaskSocket(NodeSocket):
    """Represent a socket of a Task in the AiiDA WorkGraph."""

    # use TaskProperty from aiida_workgraph.property
    # to override the default NodeProperty from node_graph
    node_property = TaskProperty

    @property
    def node_value(self):
        return self.get_node_value()

    def get_node_value(self):
        """Obtain the actual Python `value` of the object attached to the Socket."""
        if isinstance(self.value, orm.Data):
            if hasattr(self.value, "value"):
                return self.value.value
            else:
                raise ValueError(
                    "Data node does not have a value attribute. We do not know how to extract the raw Python value."
                )
        else:
            return self.value


def build_socket_from_AiiDA(DataClass: Type[Any]) -> Type[TaskSocket]:
    """Create a socket class from AiiDA DataClass."""

    class AiiDATaskSocket(TaskSocket):
        """AiiDA Task Socket."""

        identifier: str = DataClass.__name__

        def __init__(
            self,
            name: str,
            parent: Any = None,
            type: str = "INPUT",
            index: int = 0,
            uuid: str = None,
            **kwargs: Any
        ) -> None:
            super().__init__(name, parent, type, index, uuid=uuid)
            self.add_property(DataClass, name, **kwargs)

        def get_serialize(self) -> dict:
            serialize = {"module": "aiida.orm.utils.serialize", "name": "serialize"}
            return serialize

        def get_deserialize(self) -> dict:
            deserialize = {
                "module": "aiida.orm.utils.serialize",
                "name": "deserialize_unsafe",
            }
            return deserialize

    return AiiDATaskSocket
